
import random
import re
from functools import update_wrapper

from django import http
from django.conf.urls import include, url
from django.urls import reverse
from django.utils.decorators import classonlymethod
from django.utils.safestring import mark_safe

from . import actions, helpers, options, views
from .item import VersionsList

# Constant that defines a attribute points to it's parent
PARENT = "parent"
ACTION_ALIAS = "_action"


def create_new_viewclass(base, **kwargs):
    # Create a new view class based on a view instance
    data = {}
    kwargs.update(getattr(base, "changed_kwargs", {}))

    for k, v in list(kwargs.items()):
        if hasattr(base, k):
            data[k] = v

    if isinstance(base, views.CMSView):
        name = f"{base.__class__.__name__}{hex(id(base))}{random.random()}"
        parent = base.__class__
    else:
        name = base.__name__ + "Sub"
        parent = base

    return type(name, (parent,), data)


class PromiseBundle:
    def __init__(self, cls, name=None, title=None, title_plural=None):
        assert name
        self.name = name
        self.title = title
        self.title_plural = title_plural
        self.cls = cls
        self.initialized = None

    def __call__(self, child_name, parent, site):
        return self.cls(
            name=self.name,
            title=self.title,
            title_plural=self.title_plural,
            parent=parent,
            attr_on_parent=child_name,
            site=site,
        )

    @staticmethod
    def hidden_name(name):
        return f"_{name}_promise"


class URLAlias:
    """
    Alias urls to some other view or bundle. Aliases
    created in this way will not be added to the actual
    urls in the cms site. But when a url is requested
    for an attribute on a bundle that points to a URLAlias
    instance, whether that happens through a template tag
    or one of bundles view getter methods, the url or view
    returned will be the one for the aliased name/bundle.

    :param bundle_attr: The name of the bundle that this alias \
    points to. None means the current bundle, using the `PARENT` \
    constant means the view name will be looked up on the \
    parent bundle. Defaults to None.
    :param alias_to: The name of the view that you want this \
    to point to instead. Defaults to None.
    """

    def __init__(self, bundle_attr=None, alias_to=None):
        self.bundle_attr = bundle_attr
        self.alias_to = alias_to

    def get_bundle(self, current_bundle, url_kwargs, context_kwargs):
        """
        Returns the bundle to get the alias view from.
        If 'self.bundle_attr' is set, that bundle that it points to
        will be returned, otherwise the current_bundle will be
        returned.
        """
        if self.bundle_attr:
            if self.bundle_attr == PARENT:
                return current_bundle.parent

            view, name = current_bundle.get_view_and_name(self.bundle_attr)
            return view

        return current_bundle

    def get_view_name(self, requested):
        """
        Returns the name of the view to lookup.
        If `requested` is equal to 'self.bundle_attr' then
        'main' will be returned. Otherwise if `self.alias_to`
        is set the it's value will be returned. Otherwise
        the `requested` itself will be returned.

        """
        value = self.alias_to and self.alias_to or requested
        if value == self.bundle_attr:
            return "main"
        return value


class ViewAlias(URLAlias):
    """
    Works the same as URLAlias accept it allows
    you to reuse a view registered somewhere
    else as at different url on this bundle.
    """

    pass


class BundleMeta(type):
    """
    Metaclass for bundle that gathers the known views,
    subbundles and meta options from all the parent classes.
    """

    def __new__(cls, name, bases, attrs):
        meta = options.Meta()

        _children = set()
        _views = set()

        # Copy views from bases along with meta
        for base in bases[::-1]:
            val = getattr(base, "_views", None)
            if val and isinstance(val, tuple):
                _views = _views.union(set(base._views))

            val = getattr(base, "_children", None)
            if val and isinstance(val, tuple):
                _children = _children.union(set(base._children))

            if hasattr(base, "_meta"):
                meta.add_meta(base._meta)

        m = attrs.pop("Meta", None)
        meta.add_meta(m)
        for k, v in list(attrs.items()):
            if isinstance(v, PromiseBundle):
                _children.add(k)
                _views.add(k)
                attrs[v.hidden_name(k)] = v
            elif isinstance(v, views.CMSView):
                _views.add(k)
            elif isinstance(v, ViewAlias):
                _views.add(k)

        for v in _children:
            attrs.pop(v, None)

        attrs["_children"] = tuple(_children)
        attrs["_views"] = tuple(_views)
        attrs["_meta"] = meta
        cls = super().__new__(cls, name, bases, attrs)
        return cls


class Bundle(metaclass=BundleMeta):
    """
    Base bundle class. A bundle is a class that is meant to group together
    CMSViews and other bundle classes. It contains some methods to
    help the views know where to find each other, keep track of their
    url parameters and provide page navigation and headers.
    Views and sub bundles are specified as class attributes when
    creating a new Bundle class.

    Each bundle class has a options class stored at _meta. When one bundle
    inherits from another the meta class attributes are copied from all
    base classes, with the normal resolution rules applying. The exception
    is attributes containing a dictionary. In that case a copy of the
    dictionary from the further ancestor will be made and then updated
    with the dictionary from the closer. The resulting new dictionary
    is stored as the value for that attribute.

    Any time you set the value of a class attribute to the constant
    `PARENT` (also available on bundle instances as `self.parent_attr`)
    you are saying that attribute should be looked up on the parent object.
    This works for view attributes and some non view attributes like
    `navigation` and `object_header`.

    :param navigation: A list of tuples that represent the side navigation \
    items for this bundle. The format is (attribute_name, title, url_kwargs). \
    Title and url_kwargs are optional. If no title is given the title of the bundle
    that the view is on will be used. Default is an empty tuple.
    :param dashboard: A list of the tuples that represent the main navigation.\
    format is the same as `navigation`. Default is an empty tuple.
    :param required_groups: A list of groups names that a visitor must \
    be a member of to access views in this bundle. Default is an empty tuple.
    :param live_groups: A list of groups names that a visitor must \
    be a member of to access the 'live_views` in this bundle. Default is None \
    which means same as `required_groups`.
    :param object_view: The name of the view that should be rendered as \
    the object header. Defaults to 'delete'.
    :param main_list: A URLAlias for 'main' used by main views as their \
    default redirect target.

    By default the following views are created:

    * **main** - ListView
    * **add*** - FormView
    * **edit** - FormView
    * **delete** - DeleteActionView
    """

    parent_attr = PARENT
    action_alias = ACTION_ALIAS

    navigation = ()
    dashboard = ()

    required_groups = ()
    live_groups = None

    _children = ()
    _views = ()

    main = views.ListView()
    add = views.FormView(force_add=True)
    edit = views.FormView()
    delete = actions.DeleteActionView()

    main_list = URLAlias(alias_to="main")

    object_view = "delete"

    def __init__(
        self,
        title=None,
        title_plural=None,
        name=None,
        parent=None,
        attr_on_parent=None,
        site=None,
    ):
        assert name

        self.name = name
        self.title = title
        self.title_plural = title_plural
        self.admin_site = site
        self._url_params = ()
        self.attr_on_parent = attr_on_parent

        self.parent = parent
        if self.parent:
            self.name = f"{self.parent.name}_{self.name}"
            reg = rf"^{parent.get_regex_for_name(self.name, attr_on_parent)}"
            url_params = list(re.compile(reg).groupindex.keys())
            l = list(parent.url_params)
            l.extend(url_params)
            self._url_params = tuple(l)

            if self.required_groups == self.parent_attr:
                self.required_groups = self.parent.required_groups

        self.item_regex = self._meta.item_regex_base % {"name": self.name}

        # Only process defaults if we have a model
        if self._meta.model:
            if site and self._meta.primary_model_bundle:
                site.register_model(self._meta.model, self)

        added_views = []
        action_views = set(self._meta.action_views)
        for view in self._views:
            v = getattr(self, view, None)

            if v and isinstance(v, views.CMSView):
                view_kwargs = self._meta.get_kwargs_for_view(view)

                if self.live_groups and view in self._meta.live_views:
                    view_kwargs["required_groups"] = list(self.live_groups)

                setattr(self, view, create_new_viewclass(v, **view_kwargs))

                # Create aliases for action views
                if view in action_views:
                    view_name = "{0}{1}".format(view, ACTION_ALIAS)
                    if not hasattr(self, view_name):
                        setattr(self, view_name, ViewAlias(alias_to=view))
                        added_views.append(view_name)

        if added_views:
            self._views = tuple(list(self._views) + added_views)

    def set_admin_site(self, site):
        self.admin_site = site

        if site and self._meta.primary_model_bundle:
            site.register_model(self._meta.model, self)

    def _get_url_params(self):
        return self._url_params

    url_params = property(_get_url_params)

    def get_object_header_view(
        self, request, url_kwargs, parent_only=False, render_type="object_header"
    ):
        """
        An object header is the title block of a CMS page. Actions
        to linked to in the header are based on this views
        bundle.

        This returns a view instance and view name of the view that
        should be rendered as an object header the view used is specified
        in `self.object_view`. If not match is found None, None is returned

        :param request: The request object
        :param url_kwargs: Any url keyword arguments as a dictionary
        :param parent_only: If `True` then the view will only \
        be rendered if object_view points to parent. This is usually \
        what you want to avoid extra lookups to get the object \
        you already have.
        :param render_type: The render type to use for the header. \
        Defaults to 'object_header'.
        """

        if parent_only and self.object_view != self.parent_attr:
            return None, None

        if self.object_view == self.parent_attr and self.parent:
            return self.parent.get_object_header_view(
                request, url_kwargs, render_type=render_type
            )
        elif self.object_view:
            view, name = self.get_initialized_view_and_name(
                self.object_view,
                can_submit=False,
                base_template="cms/partial.html",
                request=request,
                kwargs=url_kwargs,
                render_type=render_type,
            )
            if view and view.can_view(request.user):
                return view, name
        return None, None

    def get_string_from_view(
        self, request, view_name, url_kwargs, render_type="string"
    ):

        """
        Returns a string that is a rendering of the view given a
        request, view_name, and the original url_kwargs. Makes the
        following changes the view before rendering:

        * Sets can_submit to False.
        * Adds action_url to the context. This is the url where \
        this view actually lives.
        * Sets the default base_template to be 'cms/partial.html'

        This will always call GET and never POST as any actions
        that modify data should take place on the original
        url and not like this.

        :param request: The request object.
        :param view_name: The name of the view that you want.
        :param url_kwargs: The url keyword arguments that came \
        with the request object. The view itself is responsible \
        to remove arguments that would not be part of a normal match \
        for that view. This is done by calling  the `get_url_kwargs` \
        method on the view.
        :param render_type: The render type to use. Defaults to \
        'string'.
        """

        response = ""
        try:
            view, name = self.get_initialized_view_and_name(
                view_name,
                render_type=render_type,
                can_submit=False,
                base_template="cms/partial.html",
                request=request,
                kwargs=url_kwargs,
            )

            if isinstance(view, URLAlias):
                view_name = view.get_view_name(view_name)
                bundle = view.get_bundle(self, url_kwargs, {})
                if bundle and isinstance(bundle, Bundle):
                    return bundle.get_string_from_view(
                        request, view_name, url_kwargs, render_type=render_type
                    )

            elif view:
                if view and name and view.can_view(request.user):
                    response = self._render_view_as_string(
                        view, name, request, url_kwargs
                    )
        except http.Http404:
            pass
        return response

    def _render_view_as_string(self, view, name, request, url_kwargs):
        url_kwargs = view.get_url_kwargs()
        url = reverse(f"admin:{name}", kwargs=url_kwargs)
        view.add_to_render_data(action_url=url)
        return mark_safe(view.as_string(request, **url_kwargs))

    def get_view_url(
        self,
        view_name,
        user,
        url_kwargs=None,
        context_kwargs=None,
        follow_parent=True,
        check_permissions=True,
    ):
        """
        Returns the url for a given view_name. If the view isn't
        found or the user does not have permission None is returned.
        A NoReverseMatch error may be raised if the view was unable
        to find the correct keyword arguments for the reverse function
        from the given url_kwargs and context_kwargs.

        :param view_name: The name of the view that you want.
        :param user: The user who is requesting the view
        :param url_kwargs: The url keyword arguments that came \
        with the request object. The view itself is responsible \
        to remove arguments that would not be part of a normal match \
        for that view. This is done by calling  the `get_url_kwargs` \
        method on the view.
        :param context_kwargs: Extra arguments that will be passed \
        to the view for consideration in the final keyword arguments \
        for reverse.
        :param follow_parent: If we encounter a parent reference should \
        we follow it. Defaults to True.
        :param check_permisions: Run permissions checks. Defaults to True.
        """

        view, url_name = self.get_initialized_view_and_name(
            view_name, follow_parent=follow_parent
        )

        if isinstance(view, URLAlias):
            view_name = view.get_view_name(view_name)
            bundle = view.get_bundle(self, url_kwargs, context_kwargs)

            if bundle and isinstance(bundle, Bundle):
                return bundle.get_view_url(
                    view_name,
                    user,
                    url_kwargs=url_kwargs,
                    context_kwargs=context_kwargs,
                    follow_parent=follow_parent,
                    check_permissions=check_permissions,
                )

        elif view:

            # Get kwargs from view
            if not url_kwargs:
                url_kwargs = {}

            url_kwargs = view.get_url_kwargs(context_kwargs, **url_kwargs)
            view.kwargs = url_kwargs

            if check_permissions and not view.can_view(user):
                return None

            url = reverse(f"admin:{url_name}", kwargs=url_kwargs)
            return url

    def _view_uses_name_as_url_kwarg(self, view_name):
        # Returns True if the given view_name uses
        # self.name in url kwargs
        view_name = view_name.replace(ACTION_ALIAS, "")
        return (view_name in self._meta.item_views) or (
            view_name in self._meta.action_views
        )

    def _get_slug_url_kwarg_for_name(self, view_name):
        arg = None

        if self._view_uses_name_as_url_kwarg(view_name):
            arg = f"{self.name}_pk"
        elif self.parent:
            # Get the attribute from the parent so this can be chained
            arg = self.parent._get_slug_url_kwarg_for_name(self.attr_on_parent)

        return arg

    def _get_view_kwargs(self, view, view_name):
        kwargs = {}

        if hasattr(view, "bundle"):
            kwargs["bundle"] = self

        if hasattr(view, "slug_url_kwarg"):
            kwargs["slug_url_kwarg"] = self._get_slug_url_kwarg_for_name(view_name)

        return kwargs

    def get_initialized_view_and_name(
        self, view_name, follow_parent=True, **extra_kwargs
    ):
        """
        Creates and returns a new instance of a CMSView \
        and it's url_name.

        :param view_name: The name of the view to return.
        :param follow_parent: If we encounter a parent reference should \
        we follow it. Defaults to True.
        :param extra_kwargs: Keyword arguments to pass to the view.
        """

        view, name = self.get_view_and_name(view_name)

        # Initialize the view with the right kwargs
        if hasattr(view, "as_view"):
            e = dict(extra_kwargs)
            e.update(**self._get_view_kwargs(view, view_name))
            e["name"] = view_name
            view = view(**e)

        # It is a Bundle return the main
        elif isinstance(view, Bundle):
            view, name = view.get_initialized_view_and_name("main", **extra_kwargs)
        elif view == self.parent_attr and self.parent:
            if follow_parent:
                return self.parent.get_initialized_view_and_name(
                    view_name, **extra_kwargs
                )
            else:
                view = None
                name = None
        return view, name

    def get_single_title(self):
        return self.get_title(plural=False)

    def get_title(self, plural=True):
        """
        Get's the title of the bundle. Titles can be singular
        or plural.
        """
        value = self.title
        if value == self.parent_attr:
            return self.parent.get_title(plural=plural)

        if not value and self._meta.model:
            value = helpers.model_name(
                self._meta.model,
                self._meta.custom_model_name,
                self._meta.custom_model_name_plural,
                plural,
            )
        elif self.title and plural:
            value = helpers.pluralize(self.title, self.title_plural)

        return helpers.capfirst_if_needed(value)

    def _get_bundle_from_promise(self, attname):
        assert (
            self.admin_site
        ), "You must specify an admin_site before initializing sub bundles"
        attr = f"_{attname}_bundle"
        view = getattr(self, attr, None)
        if not view:
            promise = getattr(self, PromiseBundle.hidden_name(attname), None)
            if promise:
                view = promise(attname, self, self.admin_site)
                setattr(self, attr, view)
        return view

    def get_view_and_name(self, attname):
        """
        Gets a view or bundle and returns it
        and it's url_name.
        """
        view = getattr(self, attname, None)
        if attname in self._children:
            view = self._get_bundle_from_promise(attname)

        if view:
            if attname in self._children:
                return view, view.name
            elif isinstance(view, ViewAlias):
                view_name = view.get_view_name(attname)
                bundle = view.get_bundle(self, {}, {})
                if bundle and isinstance(bundle, Bundle):
                    view, name = bundle.get_view_and_name(view_name)

            if hasattr(view, "as_view"):
                if attname != "main":
                    name = f"{self.name}_{attname}"
                else:
                    name = self.name
                return view, name
            elif view == self.parent_attr and self.parent:
                return self.parent_attr, None
            elif isinstance(view, URLAlias):
                return view, None

        return None, None

    def get_regex_for_name(self, name, attname):

        # Get the regex for this view
        regex = ""
        if name != self.name and attname != "main":
            regex = f"{attname}/"
            if hasattr(self._meta, f"{attname}_regex_base"):
                regex = getattr(self._meta, f"{attname}_regex_base")
                regex = regex % {"group_name": self.name, "attname": attname}
            elif attname in self._meta.item_views or attname in self._meta.action_views:
                regex = f"{self.item_regex}{regex}"
        return regex

    def get_url(self, name, view_obj, attname):
        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(view)(*args, **kwargs)

            return update_wrapper(wrapper, view)

        regex = self.get_regex_for_name(name, attname)
        if isinstance(view_obj, Bundle):
            reg = rf"^{regex}"
            u = url(reg, include(view_obj.get_urls()))
        else:
            view_kwargs = self._get_view_kwargs(view_obj, attname)
            u = url(rf"^{regex}$", wrap(view_obj.as_view(**view_kwargs)), name=name)
        return u

    def get_urls(self):
        """
        Returns urls handling bundles and views.
        This processes the 'item view' first in order
        and then adds any non item views at the end.
        """
        parts = []
        seen = set()

        # Process item views in order
        for v in list(self._meta.item_views) + list(self._meta.action_views):
            if v not in seen:
                view, name = self.get_view_and_name(v)
                if view and name:
                    parts.append(self.get_url(name, view, v))
                seen.add(v)

        # Process everything else that we have not seen
        for v in set(self._views).difference(seen):
            # Get the url name
            view, name = self.get_view_and_name(v)
            if view and name:
                parts.append(self.get_url(name, view, v))

        return parts

    def _optional_tuples(self, tup):
        for item in tup:
            if len(item) == 1:
                yield (item[0], None, None)
            elif len(item) == 2:
                yield (item[0], item[1], None)
            else:
                yield item

    def _nav_from_tuple(self, request, tup, **kwargs):
        navigation = []
        for view_name, title, url_kwargs in self._optional_tuples(tup):
            url = self.get_view_url(
                view_name, request.user, url_kwargs=url_kwargs, context_kwargs=kwargs
            )
            if url:
                if not title and view_name in self._children:
                    b = self._get_bundle_from_promise(view_name)
                    title = b.get_title()
                elif not title:
                    title = self.get_title()

                navigation.append((url, title))
        return navigation

    def get_dashboard_urls(self, request):
        """
        Generates a list of tuples based on the values
        in `self.dashboard` that are the main navigation links
        for this bundle. The tuple format is (url, title).
        """
        return self._nav_from_tuple(request, self.dashboard)

    def get_dashboard_block(self, request):
        """
        Returns a block of html for display on the dashboard.
        """
        return None

    def get_navigation(self, request, **kwargs):
        """
        Generates a list of tuples based on the values
        in `self.navigation` that are the side navigation links
        for this bundle. The tuple format is (url, title).
        """

        if self.navigation == self.parent_attr:
            if self.parent:
                return self.parent.get_navigation(request, **kwargs)
            return ()
        else:
            return self._nav_from_tuple(request, self.navigation, **kwargs)

    @classonlymethod
    def as_subbundle(cls, name=None, title=None, title_plural=None):
        """
        Wraps the given bundle so that it can be lazily
        instantiated.

        :param name: The slug for this bundle.
        :param title: The verbose name for this bundle.
        """
        return PromiseBundle(cls, name=name, title=title, title_plural=title_plural)


class BlankBundle(Bundle):
    """
    Base bundle that has no preset views. Should be used as a base
    for bundle's that are not meant for typical CRUD operations.
    """

    main = None
    add = None
    edit = None
    delete = None
    publish = None
    versions = None
    unpublish = None
    main_list = None


class VersionMixin:
    _views = ("publish", "unpublish", "versions")

    publish = actions.PublishActionView()
    unpublish = actions.UnPublishActionView()
    versions = VersionsList()


class VersionedBundle(Bundle, VersionMixin):
    """
    Base bundle for versioned models. Adds views for publishing,
    un-publishing and managing versions.
    """

    class Meta(options.VersionMeta):
        pass


class DelegatedObjectBundle(Bundle):
    """
    Base bundle that delegates the following views to use the
    bundle specified by edit:

    * delete
    * publish
    * unpublish
    * versions

    This is useful for bundles that contain a list but all the actions
    for items in that list are specified on the sub bundle edit.
    """

    delete = URLAlias(bundle_attr="edit")
    publish = URLAlias(bundle_attr="edit")
    unpublish = URLAlias(bundle_attr="edit")
    versions = URLAlias(bundle_attr="edit")

    delete_action = ViewAlias(bundle_attr="edit", alias_to="delete")
    publish_action = ViewAlias(bundle_attr="edit", alias_to="publish")
    unpublish_action = ViewAlias(bundle_attr="edit", alias_to="unpublish")

    class Meta(options.VersionMeta):
        pass


class ObjectOnlyBundle(Bundle):
    """
    Base Bundle for sub bundles that do not contain a list
    page. Makes the following changes

    * Removes add.
    * main is a FormView.
    * edit points to PARENT, since that is what main is.
    * main_list points to PARENT.
    * The item views attribute of meta is set to be empty.
    """

    add = None
    main = views.FormView()
    edit = PARENT

    main_list = URLAlias(bundle_attr=PARENT)
    delegated = True

    class Meta:
        item_views = ()
        action_views = ()
        live_views = ("delete", "publish", "unpublish", "versions")


class VersionedObjectOnlyBundle(ObjectOnlyBundle, VersionMixin):
    """
    Same as ObjectOnlyBundle but adds version management views.
    """

    pass


class ChildBundle(Bundle):
    """
    Base Bundle for sub bundles. Makes the following changes:

    * required_groups is inherited from PARENT.
    """

    required_groups = PARENT

    class Meta:
        pass


class ParentVersionedBundle(ChildBundle):
    """
    Same as ChildBundle expect that is also changes:

    * object_view is inherited from PARENT.
    """

    object_view = PARENT



# imports for backwards compatibility
from django.views.generic.edit import ModelFormMixin

from .actions import DeleteView, PublishView, UnPublishView
from .base_views import (
    BaseView,
    CMSView,
    ModelCMSMixin,
    ModelCMSView,
    SiteView
)
from .item import FormView, PreviewWrapper, VersionsList
from .list import ListView



from .groups import CacheGroup


class CacheManager:
    """
    A CacheManager is where you register all the cache groups
    that you are tracking.

    Similar to the django admin, most implementations would only
    have one instance of this class that all managers would be
    registered with. If you don't need any customizations you can
    simply register with the default instance
    """

    _registry = {}

    @classmethod
    def reset(cls):
        cls._registry = {}

    def get_group(self, key):
        """
        Returns the cache groups that matches the given key,
        if not such key was registered None is returned.
        """

        return self._registry.get(key)

    def register_cache(self, cache_group):
        """
        Register a cache_group with this manager.

        Use this method to register more complicated
        groups that you create yourself. Such as if you
        need to register several models each with different
        parameters.

        :param cache_group: The group to register. \
        The group is registered with the cache_group key attribute. \
        Raises an exception if the key is already registered.
        """

        if cache_group.key in self._registry:
            raise Exception(f"{cache_group.key} is already registered")
        self._registry[cache_group.key] = cache_group

    def register_model(self, key, *models, **kwargs):
        """
        Register a cache_group with this manager.

        Use this method to register more simple
        groups where all models share the same parameters.

        Any arguments are treated as models that you would like
        to register.

        Any keyword arguments received are passed to the
        register method when registering each model.

        :param key: The key to register this group as. \
        Raises an exception if the key is already registered.
        """

        assert models, "No models passed in!"
        cache_group = CacheGroup(key)
        for model in models:
            cache_group.register(model, **kwargs)

        self.register_cache(cache_group)

    def invalidate_cache(self, klass, extra=None, **kwargs):
        """
        Invalidate a cache for a specific class.

        This will loop through all registered groups that have registered
        the given model class and call their invalidate_cache method.

        All keyword arguments will be directly passed through to the
        group's invalidate_cache method, with the exception of **extra**
        as noted below.

        :param klass: The model class that need some invalidation.

        :param extra: A dictionary where the key corresponds to the name \
        of a group where this model is registered and a value that is a \
        list that will be passed as the extra keyword argument when \
        calling invalidate_cache on that group. In this way you can \
        specify specific extra values to invalidate only for specific \
        groups.
        """

        extra = extra or kwargs.pop("extra", {})
        for group in list(self._registry.values()):
            if klass in group.models:
                e = extra.get(group.key)
                group.invalidate_cache(klass, extra=e, **kwargs)


cache_manager = CacheManager()



class Meta:
    """
    The options class for Bundle objects, every bundle will
    have an instance of this class as a _meta class attribute.

    The following options, if set, are passed to all view instances
    on the bundle. For more information on what each one does
    see the CMS Views documentation.

    * model
    * parent_field
    * parent_lookups
    * base_template
    * custom_model_name
    * custom_model_name_plural

    You can specify additional arguments to all view classes by
    setting a dictionary to `default_kwargs`. You can also specify
    additional arguments to just one view class by using FOO_kwargs.

    Other settings are not passed to views. These are:

    * **item_views** - A tuple of attribute names that should be \
    treated as item views. Meaning that they need additional url \
    keyword arguments to lookup their item. The regular expression \
    to use is set by `item_regex_base` or FOO_regex_base. \
    The order of the regular expressions in the resulting url \
    config matters, so the order specified here is preserved.
    * **item_regex_base** - A regular expression string for 'item \
    views'. This argument must take the string formatter %(name)s \
    followed by '_pk' to keep it distinct from any parent url arguments. \
    Defaults to  '(?P<%(name)s_pk>\d+)/'. You can modify the regex \
    for a particular view by adding a FOO_regex_base attribute to \
    your meta class. The base regex strings should use the string \
    formatters %(group_name)s and %(attname)s. As with the regular \
    regex base, %(group_name)s must be followed by '_pk' to keep \
    it distinct from any parent url arguments.
    * **live_views** - Live views are views that will have their \
    required groups set to the live_groups attribute of the bundle. \
    Set to 'delete' by default.
    * **primary_model_bundle** Specifies that this bundle is the \
    primary bundle for it's model. This allows the custom relationship \
    widgets to be used by other CMS views that contain fields that \
    reference this model.
    """

    view_attributes = (
        "model",
        "parent_field",
        "parent_lookups",
        "base_template",
        "custom_model_name",
        "custom_model_name_plural",
    )

    other_attributes = (
        "item_regex_base",
        "item_views",
        "live_views",
        "defaults",
        "default_kwargs",
        "primary_model_bundle",
        "action_views",
    )

    def __init__(self):
        self.primary_model_bundle = False

        # which items should use item_regex_base
        self.item_views = ("edit",)

        # which items are considered live actions
        self.live_views = ("delete",)

        # which items should be displayed as mass actions
        self.action_views = ("delete",)

        # The regex that should be used to match in
        # urls the value for %s is determined by the bundle
        self.item_regex_base = "(?P<%(name)s_pk>\d+)/"

        # the models that views are based on.
        # If not give all items are ignored.
        self.model = None

        # Custom model name if you don't want to use the default
        self.custom_model_name = None
        self.custom_model_name_plural = None

        # Optional: field on model that refers to a foreign key that must
        # be present in order to work on this bundle.
        self.parent_field = None

        # Optional: Any additional fields that you need to be
        # be included when finding a parent. Only used if parent_field is
        self.parent_lookups = None

        self.add_kwargs = {"force_add": True}

        # Kwargs that get passed to all views
        self.default_kwargs = {}

    def add_meta(self, meta):
        allowed = list(self.view_attributes)
        allowed.extend(list(self.other_attributes))

        if meta:
            for k in [
                x
                for x in dir(meta)
                if x in allowed or x.endswith("_kwargs") or x.endswith("_regex_base")
            ]:

                v = getattr(meta, k)
                if isinstance(v, dict) and getattr(self, k, None):
                    tmp = dict(getattr(self, k))
                    tmp.update(v)
                    v = tmp

                setattr(self, k, v)

    def get_kwargs_for_view(self, name):
        """
        Returns the full list of keyword arguments
        for the given view name as a dictionary.
        First the default_kwargs dictionary is copied.
        Then it is updated with the any of the 'view values'
        that can be specified directly on this instance. IE: models.
        Then that dictionary is updated with the values
        particular to this view names from the FOO_kwargs dictionary.
        """

        data = dict(self.default_kwargs)
        for k in self.view_attributes:
            if hasattr(self, k):
                data[k] = getattr(self, k)
        data.update(getattr(self, f"{name}_kwargs", {}))
        return data


class VersionMeta:
    item_views = ("edit", "versions")
    live_views = ("delete", "publish", "unpublish", "versions")
    action_views = ("delete", "publish", "unpublish")


class Orderable:
    """
    Allows rows to be reordered on the 'main' list page.
    """

    main_kwargs = {
        "change_fields": ("order",),
        "base_template": "cms/base_bundle_view.html",
        "can_sort": False,
    }



from django.contrib.admin.actions import delete_selected
from django.contrib.admin.options import ModelAdmin as OrgModelAdmin
from django.contrib.admin.sites import AdminSite as OrgAdminSite
from django.contrib.auth.admin import GroupAdmin, UserAdmin
from django.contrib.auth.models import Group, User
from django.utils.translation import ugettext_lazy

from .manager import cache_manager


def clear_cache_delete_selected(modeladmin, request, queryset):
    """
    A delete action that will invalidate cache after being called.
    """
    result = delete_selected(modeladmin, request, queryset)

    # A result of None means that the delete happened.
    if not result and hasattr(modeladmin, "invalidate_cache"):
        modeladmin.invalidate_cache(queryset=queryset)

    return result


clear_cache_delete_selected.short_description = ugettext_lazy(
    "Delete selected %(verbose_name_plural)s"
)


class AdminSite(OrgAdminSite):
    """
    Admin Class that overrides the default delete action to
    invalidate appropriate cache groups.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._actions = {"delete_selected": clear_cache_delete_selected}
        self._global_actions = self._actions.copy()


class ModelAdmin(OrgModelAdmin):
    """
    ModelAdmin implementation that trys to clear caches
    based on changes you've made.

    **cache_manager:** A CacheController instance that will be used to
    invalidate caches.
    """

    cache_manager = cache_manager

    def invalidate_cache(self, obj=None, queryset=None, extra=None, force_all=False):
        """
        Method that should be called by all tiggers to invalidate the
        cache for an item(s).

        Should be overriden by inheriting classes to customize behavior.
        """

        if self.cache_manager:
            if queryset is not None:
                force_all = True

            self.cache_manager.invalidate_cache(
                self.model, instance=obj, extra=extra, force_all=force_all
            )

    def delete_model(self, request, obj):
        result = super().delete_model(request, obj)
        self.invalidate_cache(obj=obj)
        return result

    def response_change(self, request, new_object):
        result = super().response_change(request, new_object)
        self.invalidate_cache(obj=new_object)
        return result

    def response_add(self, request, obj, post_url_continue=None):
        result = super().response_add(request, obj, post_url_continue)
        self.invalidate_cache(obj=obj)
        return result


site = AdminSite()
site.register(Group, GroupAdmin)
site.register(User, UserAdmin)



from django import forms

from scarlet.cms.item import FormView
from django.forms.models import inlineformset_factory

from .models import *
import pytest


pytestmark = pytest.mark.django_db


def test_inline_model_with_to_field():
    """An inline model with a to_field of a formset with instance have working relations. Regression for #13794"""
    FormSet = inlineformset_factory(User, UserSite, fields="__all__")
    user = User.objects.create(username="guido")
    UserSite.objects.create(user=user, data=10)
    formset = FormSet(instance=user)
    formset[0]
    # Testing the inline model's relation
    assert formset[0].instance.user_id == "guido"


def test_redirects(dummy, authenticated_client):
    # Dummy_Redirector should redirect from an edit page back to the same edit page upon save.
    resp = authenticated_client.post(
        f"/admin/dummy/dummy_redirector/{dummy.pk}/edit/",
        data={"view_tags": "dummy redirects, a", "name": "B"},
    )
    assert resp.status_code == 302
    assert DummyModel.objects.filter(name="B").count() == 1
    assert DummyModel.objects.filter(name="A").count() == 0
    assert (resp["Location"])[
        resp["Location"].find("/admin/") :
    ] == f"/admin/dummy/dummy_redirector/{dummy.pk}/edit/"


def test_URLAlias(dummy, authenticated_client):
    # Dummy_Alias makes 'edit' an alias for 'dummy_edit', and all edits should be made at the latter URL
    resp = authenticated_client.post(
        f"/admin/dummy/dummy_alias/{dummy.pk}/dummy_edit/",
        data={"view_tags": "dummy aliases, b", "name": "C"},
    )
    assert resp.status_code == 302
    assert DummyModel.objects.filter(name="C").count() == 1
    assert DummyModel.objects.filter(name="D").count() == 0


def test_bundle_independence(authenticated_client):
    # test bundles that use the same subbundle have independent URLs
    resp = authenticated_client.get("/admin/dummy/author/")
    assert resp.status_code == 200

    resp = authenticated_client.get("/admin/authoronly/author/")
    assert resp.status_code == 200

    resp = authenticated_client.post(
        "/admin/dummy/author/add/",
        data={"view_tags": "authors", "name": "Two", "bio": "2"},
    )
    assert resp.status_code == 302
    assert (resp["Location"])[
        resp["Location"].find("/admin/") :
    ] == "/admin/dummy/author/"

    a = Author.objects.filter(name="Two")
    assert a.count() == 1
    resp = authenticated_client.get(f"/admin/dummy/author/{a[0].pk}/edit/")
    assert resp.status_code == 200
    resp = authenticated_client.get(f"/admin/authoronly/author/{a[0].pk}/edit/")

    resp = authenticated_client.post(
        "/admin/authoronly/author/add/",
        data={"view_tags": "authors", "name": "Three", "bio": "3"},
    )
    assert resp.status_code == 302
    assert (resp["Location"])[
        resp["Location"].find("/admin/") :
    ] == "/admin/authoronly/author/"

    a = Author.objects.filter(name="Three")
    assert a.count() == 1
    resp = authenticated_client.get(f"/admin/authoronly/author/{a[0].pk}/edit/")
    assert resp.status_code == 200
    resp = authenticated_client.get(f"/admin/dummy/author/{a[0].pk}/edit/")
    assert resp.status_code == 200


class TestUserForm(forms.ModelForm):
    class Meta:
        model = User
        fields = "__all__"


def test_wrong_fields():
    f = FormView(model=User)
    f.form = TestUserForm
    f.fieldsets = (("User", {"fields": ("first_name",)}),)
    form_class = f.get_form_class()
    form = form_class()
    assert list(form.fields.keys()), ["first_name"]



import os
import uuid

from django.core.files.uploadedfile import UploadedFile
from django.db import models
from django.forms.forms import pretty_name

from . import get_image_cropper, settings, signals, tasks, utils
from .fields import AssetRealFileField
from .managers import AssetManager

try:
    from ..versioning import manager
except ValueError:
    from versioning import manager

try:
    from ..cms.internal_tags.models import AutoTagModel
except ValueError:
    from cms.internal_tags.models import AutoTagModel


class AssetBase(AutoTagModel):
    UNKNOWN = "unknown"
    IMAGE = "image"
    DOCUMENT = "document"
    AUDIO = "audio"
    VIDEO = "video"

    TYPES = (
        settings.ASSET_TYPES
        and settings.ASSET_TYPES
        or (
            (UNKNOWN, "Unknown"),
            (IMAGE, "Image"),
            (DOCUMENT, "Document"),
            (AUDIO, "Audio"),
            (VIDEO, "Video"),
        )
    )

    __original_file = None

    title = models.CharField(max_length=255)
    file = AssetRealFileField(upload_to=utils.assets_dir)
    type = models.CharField(max_length=255, choices=TYPES, db_index=True)
    slug = models.SlugField(unique=True, max_length=255)
    user_filename = models.CharField(max_length=255)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    cbversion = models.PositiveIntegerField(editable=False)

    objects = AssetManager()

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__original_file = self.file

    def rename_file(self):
        if self.type == self.DOCUMENT:
            return False
        return settings.HASH_FILENAME

    def url(self):
        """
        This is a wrapper of file.url
        """
        return self.file.url

    def generate_slug(self):
        return str(uuid.uuid1())

    def assign_tag(self):
        pass

    def delete_real_file(self, file_obj):
        file_obj.storage.delete(file_obj.name)
        signals.file_removed.send(file_obj.name)

    def _can_crop(self):
        return self.type == self.IMAGE

    def reset_crops(self):
        """
        Reset all known crops to the default crop.

        If settings.ASSET_CELERY is specified then
        the task will be run async
        """

        if self._can_crop():
            if settings.CELERY or settings.USE_CELERY_DECORATOR:
                # this means that we are using celery
                tasks.reset_crops.apply_async(args=[self.pk], countdown=5)
            else:
                tasks.reset_crops(None, asset=self)

    def ensure_crops(self, *required_crops):
        """
        Make sure a crop exists for each crop in required_crops.
        Existing crops will not be changed.

        If settings.ASSET_CELERY is specified then
        the task will be run async
        """
        if self._can_crop():
            if settings.CELERY or settings.USE_CELERY_DECORATOR:
                # this means that we are using celery
                args = [self.pk] + list(required_crops)
                tasks.ensure_crops.apply_async(args=args, countdown=5)
            else:
                tasks.ensure_crops(None, *required_crops, asset=self)

    def create_crop(self, name, x, x2, y, y2):
        """
        Create a crop for this asset.
        """
        if self._can_crop():
            spec = get_image_cropper().create_crop(
                name, self.file, x=x, x2=x2, y=y, y2=y2
            )
            ImageDetail.save_crop_spec(self, spec)

    def save(self, *args, **kwargs):
        """
        For new assets, creates a new slug.
        For updates, deletes the old file from storage.

        Calls super to actually save the object.
        """
        if not self.pk and not self.slug:
            self.slug = self.generate_slug()

        if self.__original_file and self.file != self.__original_file:
            self.delete_real_file(self.__original_file)

        file_changed = True
        if self.pk:
            new_value = getattr(self, "file")
            if hasattr(new_value, "file"):
                file_changed = isinstance(new_value.file, UploadedFile)
        else:
            self.cbversion = 0

        if file_changed:
            self.user_filename = os.path.basename(self.file.name)
            self.cbversion = self.cbversion + 1

        if not self.title:
            self.title = self.user_filename

        super().save(*args, **kwargs)

        if file_changed:
            signals.file_saved.send(self.file.name)
            utils.update_cache_bust_version(self.file.url, self.cbversion)
            self.reset_crops()

        if self.__original_file and self.file.name != self.__original_file.name:
            with manager.SwitchSchemaManager(None):
                for related in self.__class__._meta.get_all_related_objects(
                    include_hidden=True
                ):
                    field = related.field
                    if getattr(field, "denormalize", None):
                        cname = field.get_denormalized_field_name(field.name)
                        if getattr(field, "denormalize"):
                            related.model.objects.filter(
                                **{field.name: self.pk}
                            ).update(**{cname: self.file.name})

    def delete(self, *args, **kwargs):
        """
        Deletes the actual file from storage after the object is deleted.

        Calls super to actually delete the object.
        """
        file_obj = self.file
        super().delete(*args, **kwargs)
        self.delete_real_file(file_obj)

    def __str__(self):
        return self.user_filename


class ImageDetailBase(models.Model):
    image = models.ForeignKey(settings.ASSET_MODEL, on_delete=models.CASCADE)
    width = models.PositiveIntegerField()
    height = models.PositiveIntegerField()

    name = models.CharField(max_length=255)
    editable = models.BooleanField(editable=False, default=False)

    x = models.PositiveIntegerField(null=True)
    x2 = models.PositiveIntegerField(null=True)
    y = models.PositiveIntegerField(null=True)
    y2 = models.PositiveIntegerField(null=True)

    class Meta:
        abstract = True

    def __str__(self):
        return pretty_name(self.name)

    def get_crop_config(self):
        return get_image_cropper().get_crop_config(self.name)

    @classmethod
    def save_crop_spec(cls, asset, spec, update_version=True):
        if spec:
            cdict = spec.to_dict()
            updated = cls.objects.filter(image=asset, name=cdict["name"]).update(
                **cdict
            )
            if not updated:
                cls(image=asset, **cdict).save()

            if update_version:
                asset.__class__.objects.filter(pk=asset.pk).update(
                    cbversion=models.F("cbversion") + 1
                )


class Asset(AssetBase):
    class Meta:
        abstract = False


class ImageDetail(ImageDetailBase):
    class Meta:
        abstract = False



default_app_config = "scarlet.cms.apps.AppConfig"

try:

    def autodiscover():
        from django.utils.module_loading import autodiscover_modules

        from .sites import site

        autodiscover_modules("cms_bundles", register_to=site)


except ImportError:

    def autodiscover():
        """
        Copied from django source

        Auto-discover INSTALLED_APPS cms.py modules and fail silently when
        not present. This forces an import on them to register any admin bits they
        may want.
        """

        import copy

        from django.conf import settings
        from django.utils.importlib import import_module
        from django.utils.module_loading import module_has_submodule

        from .sites import site

        for app in settings.INSTALLED_APPS:
            mod = import_module(app)
            # Attempt to import the app's admin module.
            try:
                before_import_registry = copy.copy(site._registry)
                import_module(f"{app}.cms_bundles")
            except Exception as e:
                # Reset the model registry to the state before the last import as
                # this import will have to reoccur on the next request and this
                # could raise NotRegistered and AlreadyRegistered exceptions
                # (see #8245).
                site._registry = before_import_registry

                # Decide whether to bubble up this error. If the app just
                # doesn't have an admin module, we can ignore the error
                # attempting to import it, otherwise we want it to bubble up.
                if module_has_submodule(mod, "cms_bundles"):
                    raise



from django.contrib.auth import get_user_model
from django.contrib.auth.forms import AdminPasswordChangeForm

try:
    from ..cms import bundles, views
    from ..cms.sites import site
except ValueError:
    from cms import bundles, views
    from cms.sites import site

from . import forms, groups


class AddView(views.FormView):
    force_add = True
    form_class = forms.SignupModelForm

    fieldsets = (
        (
            "User Information",
            {"fields": ("username", "first_name", "last_name", "email")},
        ),
        ("Password", {"fields": ("password1", "password2")}),
    )


class PasswordView(views.FormView):
    redirect_to_view = "edit"
    context_object_name = "object"

    def get_form_class(self):
        return AdminPasswordChangeForm

    def write_message(self, message=None):
        message = f"{self.object} password changed"
        super().write_message(message=message)

    def get_form_kwargs(self):
        # Since we aren't using a model form
        # strip instance and use user instead
        kwargs = super().get_form_kwargs()
        instance = kwargs.pop("instance")
        kwargs["user"] = instance
        return kwargs


class AccountBundle(bundles.Bundle):
    required_groups = (groups.ADMIN,)

    add = AddView()
    edit = views.FormView(
        form_class=forms.UserForm,
        fieldsets=(
            (
                "User Information",
                {
                    "fields": (
                        "username",
                        "first_name",
                        "last_name",
                        "email",
                        "password",
                    )
                },
            ),
            ("Status", {"fields": ("is_active", "is_superuser", "is_staff")}),
            ("Groups", {"fields": ("groups",)}),
        ),
        context_object_name="object",
    )
    password = PasswordView()
    main = views.ListView(
        paginate_by=100,
        display_fields=("username", "first_name", "last_name", "email", "groups"),
        action_links=(
            ("edit", "Edit", "e"),
            ("delete", "Delete", "d"),
            ("password", "Change Password", "k"),
        ),
    )

    class Meta:
        model = get_user_model()
        primary_model_bundle = True
        item_views = ("password", "edit", "delete")
        default_kwargs = {"object_header_tmpl": "cms/object_header_no_preview.html"}


site.register("users", AccountBundle(name="accounts_admin", title="Account"), 10)



# Generated by Django 2.2.11 on 2020-03-30 22:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0002_auto_20150507_1728'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cmslog',
            name='action',
            field=models.PositiveIntegerField(choices=[(0, 'Save'), (2, 'Delete'), (3, 'Published'), (4, 'Unpublished'), (5, 'Scheduled')]),
        ),
    ]



# Generated by Django 1.9 on 2016-02-04 00:36

from .. import models as cms_bundles_models
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import scarlet.cms.fields
import scarlet.versioning.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Category",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_save", models.DateTimeField(editable=False)),
                ("is_published", models.BooleanField(default=False, editable=False)),
                (
                    "created_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("v_last_save", models.DateTimeField(editable=False, null=True)),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("published", "published"),
                            ("scheduled", "scheduled"),
                            ("draft", "draft"),
                            ("archived", "archived"),
                        ],
                        editable=False,
                        max_length=50,
                    ),
                ),
                ("last_scheduled", models.DateTimeField(editable=False, null=True)),
                ("date_published", models.DateTimeField(editable=False, null=True)),
                (
                    "user_published",
                    models.CharField(editable=False, max_length=255, null=True),
                ),
                ("vid", models.PositiveIntegerField(editable=False, unique=True)),
                ("object_id", models.PositiveIntegerField(editable=False)),
                ("category", models.CharField(max_length=150)),
                ("slug", models.SlugField(editable=False, max_length=150)),
            ],
            options={"db_table": "cms_bundles_category_base", "managed": False,},
            bases=(models.Model, cms_bundles_models.CategoryReferences),
        ),
        migrations.CreateModel(
            name="Comment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_save", models.DateTimeField(editable=False)),
                ("is_published", models.BooleanField(default=False, editable=False)),
                (
                    "created_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("v_last_save", models.DateTimeField(editable=False, null=True)),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("published", "published"),
                            ("scheduled", "scheduled"),
                            ("draft", "draft"),
                            ("archived", "archived"),
                        ],
                        editable=False,
                        max_length=50,
                    ),
                ),
                ("last_scheduled", models.DateTimeField(editable=False, null=True)),
                ("date_published", models.DateTimeField(editable=False, null=True)),
                (
                    "user_published",
                    models.CharField(editable=False, max_length=255, null=True),
                ),
                ("vid", models.PositiveIntegerField(editable=False, unique=True)),
                ("object_id", models.PositiveIntegerField(editable=False)),
                ("name", models.CharField(max_length=100)),
                ("text", models.TextField()),
            ],
            options={"db_table": "cms_bundles_comment_base", "managed": False,},
            bases=(models.Model, cms_bundles_models.CommentReferences),
        ),
        migrations.CreateModel(
            name="Post",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_save", models.DateTimeField(editable=False)),
                ("is_published", models.BooleanField(default=False, editable=False)),
                (
                    "created_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("v_last_save", models.DateTimeField(editable=False, null=True)),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("published", "published"),
                            ("scheduled", "scheduled"),
                            ("draft", "draft"),
                            ("archived", "archived"),
                        ],
                        editable=False,
                        max_length=50,
                    ),
                ),
                ("last_scheduled", models.DateTimeField(editable=False, null=True)),
                ("date_published", models.DateTimeField(editable=False, null=True)),
                (
                    "user_published",
                    models.CharField(editable=False, max_length=255, null=True),
                ),
                ("vid", models.PositiveIntegerField(editable=False, unique=True)),
                ("object_id", models.PositiveIntegerField(editable=False)),
                ("date", models.DateField()),
                ("title", models.CharField(max_length=255)),
                ("slug", models.SlugField(editable=False, max_length=255)),
                ("body", models.TextField()),
                ("keywords", models.TextField(blank=True)),
                ("description", models.TextField(blank=True)),
            ],
            options={"db_table": "cms_bundles_post_base", "managed": False,},
            bases=(models.Model, cms_bundles_models.PostReferences),
        ),
        migrations.CreateModel(
            name="Author",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("bio", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="Category_base",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_published", models.BooleanField(default=False, editable=False)),
                (
                    "created_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("v_last_save", models.DateTimeField(editable=False, null=True)),
            ],
            options={"abstract": False,},
        ),
        migrations.CreateModel(
            name="Category_version",
            fields=[
                ("last_save", models.DateTimeField(editable=False)),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("published", "published"),
                            ("scheduled", "scheduled"),
                            ("draft", "draft"),
                            ("archived", "archived"),
                        ],
                        editable=False,
                        max_length=50,
                    ),
                ),
                ("last_scheduled", models.DateTimeField(editable=False, null=True)),
                ("date_published", models.DateTimeField(editable=False, null=True)),
                (
                    "user_published",
                    models.CharField(editable=False, max_length=255, null=True),
                ),
                ("category", models.CharField(max_length=150)),
                ("slug", models.SlugField(editable=False, max_length=150)),
                ("vid", models.AutoField(primary_key=True, serialize=False)),
                (
                    "object",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="version_version",
                        to="cms_bundles.Category_base",
                    ),
                ),
            ],
            options={"managed": True,},
            bases=(models.Model, cms_bundles_models.CategoryVersionReferences),
        ),
        migrations.CreateModel(
            name="Comment_base",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_published", models.BooleanField(default=False, editable=False)),
                (
                    "created_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("v_last_save", models.DateTimeField(editable=False, null=True)),
            ],
            options={"abstract": False,},
        ),
        migrations.CreateModel(
            name="Comment_version",
            fields=[
                ("last_save", models.DateTimeField(editable=False)),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("published", "published"),
                            ("scheduled", "scheduled"),
                            ("draft", "draft"),
                            ("archived", "archived"),
                        ],
                        editable=False,
                        max_length=50,
                    ),
                ),
                ("last_scheduled", models.DateTimeField(editable=False, null=True)),
                ("date_published", models.DateTimeField(editable=False, null=True)),
                (
                    "user_published",
                    models.CharField(editable=False, max_length=255, null=True),
                ),
                ("name", models.CharField(max_length=100)),
                ("text", models.TextField()),
                ("vid", models.AutoField(primary_key=True, serialize=False)),
                (
                    "object",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="version_version",
                        to="cms_bundles.Comment_base",
                    ),
                ),
                (
                    "post",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms_bundles.Post",
                    ),
                ),
            ],
            options={"managed": True,},
            bases=(models.Model, cms_bundles_models.CommentVersionReferences),
        ),
        migrations.CreateModel(
            name="DummyModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="Post_base",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_published", models.BooleanField(default=False, editable=False)),
                (
                    "created_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("v_last_save", models.DateTimeField(editable=False, null=True)),
            ],
            options={"abstract": False,},
        ),
        migrations.CreateModel(
            name="Post_version",
            fields=[
                ("last_save", models.DateTimeField(editable=False)),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("published", "published"),
                            ("scheduled", "scheduled"),
                            ("draft", "draft"),
                            ("archived", "archived"),
                        ],
                        editable=False,
                        max_length=50,
                    ),
                ),
                ("last_scheduled", models.DateTimeField(editable=False, null=True)),
                ("date_published", models.DateTimeField(editable=False, null=True)),
                (
                    "user_published",
                    models.CharField(editable=False, max_length=255, null=True),
                ),
                ("date", models.DateField()),
                ("title", models.CharField(max_length=255)),
                ("slug", models.SlugField(editable=False, max_length=255)),
                ("body", models.TextField()),
                ("keywords", models.TextField(blank=True)),
                ("description", models.TextField(blank=True)),
                ("vid", models.AutoField(primary_key=True, serialize=False)),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="cms_bundles.Author",
                    ),
                ),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms_bundles.Category",
                    ),
                ),
                (
                    "object",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="version_version",
                        to="cms_bundles.Post_base",
                    ),
                ),
            ],
            options={"managed": True,},
            bases=(models.Model, cms_bundles_models.PostVersionReferences),
        ),
        migrations.CreateModel(
            name="PostImage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_save", models.DateTimeField(editable=False)),
                ("caption", models.CharField(blank=True, max_length=255)),
                ("order", scarlet.cms.fields.OrderField(db_index=True, default=0)),
                (
                    "post",
                    scarlet.versioning.fields.FKToVersion(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms_bundles.Post_version",
                    ),
                ),
            ],
            options={"abstract": False,},
        ),
        migrations.CreateModel(
            name="Tag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="UserSite",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("data", models.IntegerField()),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                        to_field="username",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="post_version",
            name="tags",
            field=scarlet.versioning.fields.M2MFromVersion(
                blank=True, to="cms_bundles.Tag"
            ),
        ),
    ]



import json

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models


class CastOnAssignDescriptor:
    """
    A property descriptor which ensures that `field.to_python()` is called on _every_ assignment to the field.
    This used to be provided by the `django.db.models.subclassing.Creator` class, which in turn
    was used by the deprecated-in-Django-1.10 `SubfieldBase` class, hence the reimplementation here.
    """

    def __init__(self, field):
        self.field = field

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        return obj.__dict__[self.field.name]

    def __set__(self, obj, value):
        obj.__dict__[self.field.name] = self.field.to_python(value)


class JSONField(models.TextField):
    # Used so to_python() is called
    def __init__(self, *args, **kwargs):
        self.dump_kwargs = kwargs.pop("dump_kwargs", {"cls": DjangoJSONEncoder})
        self.load_kwargs = kwargs.pop("load_kwargs", {})

        super().__init__(*args, **kwargs)

    def to_python(self, value):
        if value is None or value == "":
            return {}
        elif isinstance(value, str):
            return json.loads(value, **self.load_kwargs)
        else:
            return value

    def get_db_prep_value(self, value, connection, prepared=False):
        """Convert JSON object to a string"""
        if isinstance(value, str):
            return value
        return json.dumps(value, **self.dump_kwargs)

    def contribute_to_class(self, cls, name):
        super().contribute_to_class(cls, name)
        setattr(cls, name, CastOnAssignDescriptor(self))



from django import forms
from django.db import models

from . import widgets


class OrderFormField(forms.IntegerField):
    """
    Form field for order fields.

    :param is_order_field: Set to True
    """

    is_order_field = True


class OrderField(models.PositiveIntegerField):
    """
    PositiveIntegerField that should be used to order the model
    it appears in. Default is set to 0.

    When this field is added to a model that does not specify
    default ordering on its meta class, ordering will be set to
    this field.

    Uses `OrderFormField` as its default form field.
    """

    def __init__(self, *args, **kwargs):
        if "default" not in kwargs:
            kwargs["default"] = 0
        kwargs["db_index"] = True
        super().__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name):
        super().contribute_to_class(cls, name)
        if not cls._meta.ordering:
            cls._meta.ordering = ("order",)

    def formfield(self, form_class=OrderFormField, **kwargs):
        return super().formfield(form_class=form_class, **kwargs)


class HTMLTextField(models.TextField):
    """
    TextField that uses a WYSIWYG editor as its default widget.
    """

    def formfield(self, *args, **kwargs):
        if kwargs.get("widget", None) is None:
            kwargs["widget"] = widgets.HTMLWidget
        return super().formfield(*args, **kwargs)


class AnnotationHTMLTextField(HTMLTextField):
    """
    TextField that uses a WYSIWYG editor with annotation support as its
    default widget.
    """

    def formfield(self, *args, **kwargs):
        kwargs["widget"] = widgets.AnnotatedHTMLWidget
        return super(HTMLTextField, self).formfield(*args, **kwargs)



from django.db import models
from django.db.models.fields import related

from .models import VersionView


class FKToVersion(models.ForeignKey):
    """
    Field that creates a relation between a
    version and another model
    """

    def __init__(self, *args, **kwargs):
        kwargs["to_field"] = "vid"
        # Two cases that should only be caused from an upgrade
        # of an old project where certain params weren't required
        if kwargs.get("on_delete"):
            on_delete = kwargs.get("on_delete")
            del kwargs["on_delete"]
        else:
            on_delete = models.CASCADE

        if kwargs.get("to"):
            to = kwargs.get("to")
            del kwargs["to"]
        else:
            to = args[0]

        args = (to, on_delete)

        super().__init__(*args, **kwargs)

    def deconstruct(self):
        """
        FK to version always points to a version table
        """
        name, path, args, kwargs = super().deconstruct()
        if not kwargs["to"].endswith("_version"):
            kwargs["to"] = "{0}_version".format(kwargs["to"])
        return name, path, args, kwargs


class M2MFromVersion(models.ManyToManyField):
    """
    Field that creates a many to many relation between a
    version and another model.
    """

    def __init__(self, to, **kwargs):
        # Symmetrical doesn't work with M2m relationships to
        # self and versioning.

        if to == "self":
            kwargs["symmetrical"] = False

        super().__init__(to, **kwargs)

    def update_rel_to(self, klass):
        """
        If we have a string for a model, see if we know about it yet,
        if so use it directly otherwise take the lazy approach.
        This check is needed because this is called before
        the main M2M field contribute to class is called.
        """
        if isinstance(related.resolve_relation(klass, self.remote_field.model), str):
            relation = related.resolve_relation(klass, self.remote_field.model)
            try:
                app_label, model_name = relation.split(".")
            except ValueError:
                # If we can't split, assume a model in current app
                app_label = klass._meta.app_label
                model_name = relation

            model = None
            try:
                model = klass._meta.apps.get_registered_model(app_label, model_name)
            # For django < 1.6
            except AttributeError:
                model = models.get_model(
                    app_label, model_name, seed_cache=False, only_installed=False
                )
            except LookupError:
                print(
                    f"LookupError: Unable to find model {app_label}.{model_name}."
                )

            if model:
                self.remote_field.model = model

    def contribute_to_class(self, cls, name):
        """
        Because django doesn't give us a nice way to provide
        a through table without losing functionality. We have to
        provide our own through table creation that uses the
        FKToVersion field to be used for the from field.
        """

        self.update_rel_to(cls)

        # Called to get a name
        self.set_attributes_from_name(name)
        self.model = cls

        # Set the through field
        if not self.remote_field.through and not cls._meta.abstract:
            self.remote_field.through = create_many_to_many_intermediary_model(
                self, cls
            )

        # Do the rest
        super().contribute_to_class(cls, name)


def create_many_to_many_intermediary_model(field, klass):
    """
    Copied from django, but uses FKToVersion for the
    'from' field. Fields are also always called 'from' and 'to'
    to avoid problems between version combined models.
    """
    managed = True
    temp_to_model = related.resolve_relation(klass, field.remote_field.model)
    if (
        isinstance(temp_to_model, str)
        and temp_to_model != related.RECURSIVE_RELATIONSHIP_CONSTANT
    ):
        to_model = temp_to_model
        to = to_model.split(".")[-1]

        def set_managed(model, related, through):
            through._meta.managed = model._meta.managed or related._meta.managed

        lazy_name = f"{klass._meta.object_name}_{field.name}"
        related.lazy_related_operation(set_managed, klass, to_model, lazy_name)
    elif isinstance(temp_to_model, str):
        to = klass._meta.object_name
        to_model = klass
        managed = klass._meta.managed
    else:
        to = temp_to_model._meta.object_name
        to_model = temp_to_model
        managed = klass._meta.managed or to_model._meta.managed
        if issubclass(klass, VersionView):
            managed = False

    name = f"{klass._meta.object_name}_{field.name}"
    if (
        temp_to_model == related.RECURSIVE_RELATIONSHIP_CONSTANT
        or to == klass._meta.object_name
    ):
        from_ = f"from_{to.lower()}"
        to = f"to_{to.lower()}"
    else:
        from_ = klass._meta.object_name.lower()
        to = to.lower()

    meta = type(
        "Meta",
        (object,),
        {
            "db_table": field._get_m2m_db_table(klass._meta),
            "managed": managed,
            "auto_created": klass,
            "app_label": klass._meta.app_label,
            "db_tablespace": klass._meta.db_tablespace,
            "unique_together": ("from", "to"),
            "verbose_name": f"{from_}-{to} relationship",
            "verbose_name_plural": f"{from_}-{to} relationships",
            "apps": field.model._meta.apps,
        },
    )

    # Construct and return the new class.
    return type(
        str(name),
        (models.Model,),
        {
            "Meta": meta,
            "__module__": klass.__module__,
            "from": FKToVersion(
                klass,
                related_name=f"{name}+",
                db_tablespace=field.db_tablespace,
                db_constraint=field.remote_field.db_constraint,
                on_delete=models.CASCADE,
            ),
            "to": models.ForeignKey(
                to_model,
                related_name=f"{name}+",
                db_tablespace=field.db_tablespace,
                db_constraint=field.remote_field.db_constraint,
                on_delete=models.CASCADE,
            ),
        },
    )



from django import forms

from .models import *


class EditAuthorForm(forms.ModelForm):
    """
    Form for handling asset updates
    """

    class Meta:
        model = Author
        fields = ("name", "bio")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import copy
import urllib.parse
from functools import update_wrapper

from django import http
from django import VERSION as DJANGO_VERSION
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.decorators import classonlymethod, method_decorator
from django.utils.encoding import force_text
from django.views import generic
from django.views.decorators.csrf import csrf_protect

from . import helpers, renders, widgets
from .internal_tags import handler as tag_handler
from .models import CMSLog


class BaseView(generic.base.View):
    """
    Base view for all Views. This is a class based view
    that uses render classes to handle the preparation of an
    response.

    Render classes are classes that return a response of
    some kind from a view, such as a HttpResponse object or a string.
    The only requirement is that they have a render method that takes
    a request and keyword arguments. Those keyword arguments should be
    considered the context.

    :param renders: A dictionary where the values are render \
    class instances.
    :param render_type: A string that denotes what renderer \
    should be used. This must but a key in the renders dictionary. \
    Default is 'response'.
    """

    render_type = "response"

    def get_render_data(self, **kwargs):
        """
        Because of the way mixin inheritance works
        we can't have a default implementation of
        get_context_data on the this class, so this
        calls that method if available and returns
        the resulting context.
        """
        if hasattr(self, "get_context_data"):
            data = self.get_context_data(**kwargs)
        else:
            data = kwargs
        return data

    def render(self, request, collect_render_data=True, **kwargs):
        """
        Render this view. This will call the render method
        on the render class specified.

        :param request: The request object
        :param collect_render_data: If True we will call \
        the get_render_data method to pass a complete context \
        to the renderer.
        :param kwargs: Any other keyword arguments that should \
        be passed to the renderer.
        """
        assert self.render_type in self.renders
        render = self.renders[self.render_type]
        if collect_render_data:
            kwargs = self.get_render_data(**kwargs)

        return render.render(request, **kwargs)


class SiteView(BaseView):
    """
    Base view that provides a classmethod to render
    the view as a string instead of as a HttpResponse.
    """

    def get_as_string(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    @classonlymethod
    def as_string(cls, **initkwargs):
        """
        Similar to the as_view classmethod except this method will
        render this view as a string. When rendering a view this way
        the request will always be routed to the get method.
        The default render_type is 'string' unless you specify
        something else. If you provide your own render_type be sure
        to specify a render class that returns a string.
        """

        if "render_type" not in initkwargs:
            initkwargs["render_type"] = "string"

        for key in initkwargs:
            if key in cls.http_method_names:
                raise TypeError(
                    f"You tried to pass in the {key} method name as a"
                    f" keyword argument to {cls.__name__}(). Don't do that."
                )
            if not hasattr(cls, key):
                raise TypeError(
                    f"{cls.__name__}() received an invalid keyword {key}"
                )

        def view(request, *args, **kwargs):
            try:
                self = cls(**initkwargs)
                self.request = request
                self.args = args
                self.kwargs = kwargs
                return self.get_as_string(request, *args, **kwargs)
            except http.Http404:
                return ""

        # take name and docstring from class
        update_wrapper(view, cls, updated=())

        return view


class CMSView(BaseView):
    """
    Base class for CMS Views, should be used as the main view
    class for all views that will be registered with a bundle.

    :param base_template: The default base template, gets passed \
    to the main template as 'base'.
    :param default_template: The main template for this view.
    :param object_header_tmpl: The template to render for an object \
    header.
    :param render_type: The render_type, defaults to 'response'.
    :param renders: The dictionary of render views that this view \
    knows how to set. Unless specified otherwise, this will be set \
    to a CMSRender instance 'response' and RenderString instance \
    'string'; both instances are initialized with the `default_template`.
    :param bundle: The bundle that this view instance is attached \
    to. Set by the bundle on initialization.
    :param can_submit: Does this instance support form submission. \
    Set to false by bundle when necessary.
    :param required_groups: Groups that are required in order to render \
    this view. Defaults to None, may be set by bundle.
    """

    object_header_tmpl = "cms/object_header.html"

    render_type = "response"

    base_template = "cms/base_bundle_view.html"

    renders = {}

    bundle = None

    can_submit = True

    required_groups = None

    ORIGIN_ARGUMENT = "o"

    name = None

    def __init__(self, *args, **kwargs):
        self.extra_render_data = {}
        self.changed_kwargs = kwargs
        super().__init__(*args, **kwargs)
        if not self.renders:
            self.renders = {
                "response": renders.CMSRender(
                    template=self.default_template, base=self.base_template
                ),
                "string": renders.RenderString(
                    template=self.default_template, base=self.base_template
                ),
            }

    def customized_return_url(self, default_url):
        redirect_url = self.request.GET.get(self.ORIGIN_ARGUMENT)
        if redirect_url:
            base = self.request.path.split("/")
            if len(base) > 1 and redirect_url.startswith(f"/{base[1]}"):
                return redirect_url

        return default_url

    def get_navigation(self):
        """
        Hook for overiding navigation per view.
        Defaults to calling get_navigation on the
        bundle.
        """
        return self.bundle.get_navigation(self.request, **self.kwargs)

    def add_to_render_data(self, **kwargs):
        """
        Any keyword arguments provided will be passed to
        the renderer when this view is rendered.
        """
        self.extra_render_data.update(kwargs)

    def get_tags(self, view_object=None):
        """
        This method return a list of tags to use in the template
        :return: list of tags
        """
        tags = [force_text(self.bundle.get_title())]
        back_bundle = self.get_back_bundle()
        if back_bundle and back_bundle != self.bundle:
            tags.append(force_text(back_bundle.get_title()))
        if view_object:
            tags.append(force_text(view_object))

        return tags

    def get_back_bundle(self, start_bundle=None):
        if not start_bundle:
            start_bundle = self.bundle

        if getattr(start_bundle, "main_list", None):
            main_list = start_bundle.main_list
            bundle = main_list.get_bundle(start_bundle, {}, self.kwargs)
            return bundle
        return None

    def get_render_data(self, **kwargs):
        """
        Returns all data that should be passed to the renderer.
        By default adds the following arguments:

        * **bundle** - The bundle that is attached to this view instance.
        * **url_params** - The url keyword arguments. i.e.: self.kwargs.
        * **user** - The user attached to this request.
        * **base** - Unless base was already specified this gets set to \
        'self.base_template'.
        * **navigation** - The navigation bar for the page
        * **object_header_tmpl** - The template to use for the \
        object_header. Set to `self.object_header_tmpl`.
        * **back_bundle** - The back_back bundle is bundle that is linked to \
        from the object header as part of navigation. If there is an 'obj' \
        argument in the context to render, this will be set to the bundle \
        pointed to by the `main_list` attribute of this view's bundle. \
        If this is not set, the template's back link will point to the \
        admin_site's home page.
        """

        obj = getattr(self, "object", None)
        data = dict(self.extra_render_data)
        data.update(kwargs)
        data.update(
            {
                "bundle": self.bundle,
                "navigation": self.get_navigation(),
                "url_params": self.kwargs,
                "user": self.request.user,
                "object_header_tmpl": self.object_header_tmpl,
                "view_tags": tag_handler.tags_to_string(self.get_tags(obj)),
            }
        )

        if "base" not in data:
            data["base"] = self.base_template

        if "back_bundle" not in data:
            data["back_bundle"] = self.get_back_bundle()

        return super().get_render_data(**data)

    def _user_in_groups(self, user, allowed_groups):
        groups = getattr(user, "cached_groups", None)
        if groups is None:
            user.cached_groups = user.groups.all()

        for group in user.cached_groups:
            if group.name in allowed_groups:
                return True
        return False

    def can_view(self, user):
        """
        Returns True if user has permission to render this view.

        At minimum this requires an active staff user. If the required_groups
        attribute is not empty then the user must be a member of at least one
        of those groups. If there are no required groups set for the view but
        required groups are set for the bundle then the user must be a member
        of at least one of those groups. If there are no groups to check this
        will return True.
        """

        if user.is_staff and user.is_active:
            if user.is_superuser:
                return True
            elif self.required_groups:
                return self._user_in_groups(user, self.required_groups)
            elif self.bundle.required_groups:
                return self._user_in_groups(user, self.bundle.required_groups)
            else:
                return True

        return False

    def get_url_kwargs(self, request_kwargs=None, **kwargs):
        """
        Get the kwargs needed to reverse this url.

        :param request_kwargs: The kwargs from the current request. \
        These keyword arguments are only retained if they are present \
        in this bundle's known url_parameters.
        :param kwargs: Keyword arguments that will always be kept.
        """

        if not request_kwargs:
            request_kwargs = getattr(self, "kwargs", {})

        for k in self.bundle.url_params:
            if k in request_kwargs and k not in kwargs:
                kwargs[k] = request_kwargs[k]
        return kwargs

    def customize_form_widgets(self, form_class, fields=None):
        """
        Hook for customizing widgets for a form_class. This is needed
        for forms that specify their own fields causing the
        default db_field callback to not be run for that field.

        Default implementation checks for APIModelChoiceWidgets
        or APIManyChoiceWidgets and runs the update_links method
        on them. Passing the admin_site and request being used.

        Returns a new class that contains the field with the initialized
        custom widget.
        """
        attrs = {}
        if fields:
            fields = set(fields)

        for k, f in list(form_class.base_fields.items()):
            if fields and k not in fields:
                continue

            if isinstance(f.widget, widgets.APIModelChoiceWidget) or isinstance(
                f.widget, widgets.APIManyChoiceWidget
            ):
                field = copy.deepcopy(f)
                field.widget.update_links(self.request, self.bundle.admin_site)
                attrs[k] = field

        if attrs:
            form_class = type(form_class.__name__, (form_class,), attrs)

        return form_class

    @method_decorator(csrf_protect)
    def dispatch(self, request, *args, **kwargs):
        """
        Overrides the custom dispatch method to raise a Http404
        if the current user does not have view permissions.
        """
        self.request = request
        self.args = args
        self.kwargs = kwargs

        if not self.can_view(request.user):
            raise http.Http404

        return super().dispatch(request, *args, **kwargs)

    def as_string(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)


class ModelCMSMixin:
    """
    Mixin for use with all cms views that interact with database models.

    :param parent_field: The name of the field on the model \
    that points a parent that must be present. This should either be \
    a foreign key or a many to many field. Defaults to None.
    :param parent_lookups: A list of field names to use when constructing \
    the query for the parent object. If `parent_field` is present defaults \
    to ('pk',) otherwise default is None.
    :param slug_field: The name of the field on the model that should \
    be used as the keyword argument when looking up an object. Used in \
    combination with `slug_url_kwarg`. Defaults to 'pk'
    :param slug_url_kwarg: The name of the URLConf keyword argument \
    that contains the value to filter on. Used in combination with \
    `slug_field`. Defaults to 'pk'.
    :param base_filter_kwargs: Attribute to provide a dictionary of \
    default keyword arguments for the base queryset.
    """

    parent_lookups = None
    parent_field = None

    slug_field = "pk"
    slug_url_kwarg = "pk"

    base_filter_kwargs = {}

    def get_formfield_overrides(self):
        """
        Hook for specifying the default arguments when
        creating a form field for a db field.

        By default adds DateWidget for date fields,
        TimeChoiceWidget for time fields, SplitDateTime
        for datetime fields, and APIChoiceWidget for
        foreign keys.
        """
        return helpers.FORMFIELD_FOR_DBFIELD_DEFAULTS

    def formfield_for_dbfield(self, db_field, **kwargs):
        """
        Hook for specifying the form Field instance for a given
        database Field instance. If kwargs are given, they're
        passed to the form Field's constructor.

        Default implementation uses the overrides returned by
        `get_formfield_overrides`. If a widget is an instance
        of APIChoiceWidget this will do lookup on the current
        admin site for the bundle that is registered for that
        module as the primary bundle for that one model. If a
        match is found then this will call update_links on that
        widget to store the appropriate urls for the javascript
        to call. Otherwise the widget is removed and the default
        select widget will be used instead.
        """
        overides = self.get_formfield_overrides()

        # If we've got overrides for the formfield defined, use 'em. **kwargs
        # passed to formfield_for_dbfield override the defaults.
        for klass in db_field.__class__.mro():
            if klass in overides:
                kwargs = dict(overides[klass], **kwargs)
                break

        # Our custom widgets need special init
        mbundle = None
        extra = kwargs.pop("widget_kwargs", {})
        widget = kwargs.get("widget")
        if widget:
            if (
                widget
                and isinstance(widget, type)
                and issubclass(widget, widgets.APIChoiceWidget)
            ):
                mbundle = self.bundle.admin_site.get_bundle_for_model(
                    db_field.remote_field.model
                )
                if mbundle:
                    widget = widget(db_field.remote_field, **extra)
                else:
                    widget = None

        if (
            getattr(self, "prepopulated_fields", None)
            and not getattr(self, "object", None)
            and db_field.name in self.prepopulated_fields
        ):
            extra = kwargs.pop("widget_kwargs", {})
            attr = extra.pop("attrs", {})
            attr["data-source-fields"] = self.prepopulated_fields[db_field.name]
            extra["attrs"] = attr
            if not widget:
                from django.forms.widgets import TextInput

                widget = TextInput(**extra)
            elif widget and isinstance(widget, type):
                widget = widget(**extra)

        kwargs["widget"] = widget

        field = db_field.formfield(**kwargs)
        if mbundle:
            field.widget.update_links(self.request, self.bundle.admin_site)
        return field

    def log_action(
        self, instance, action, action_date=None, url="", update_parent=True
    ):
        """
        Store an action in the database using the CMSLog model.
        The following attributes are calculated and set on the log entry:

         * **model_repr** - A unicode representation of the instance.
         * **object_repr** - The verbose_name of the instance model class.
         * **section** - The name of ancestor bundle that is directly \
         attached to the admin site.

        :param instance: The instance that this action was performed \
        on.
        :param action: The action type. Must be one of the options \
        in CMSLog.ACTIONS.
        :param action_date: The datetime the action occurred.
        :param url: The url that the log entry should point to, \
        Defaults to an empty string.
        :param update_parent: If true this will update the last saved time \
        on the object pointed to by this bundle's object_view. \
        Defaults to True.
        """

        section = None
        if self.bundle:
            bundle = self.bundle
            while bundle.parent:
                bundle = bundle.parent
            section = bundle.name

        # if we have a object view that comes from somewhere else
        # save it too to update it.
        changed_object = instance
        bundle = self.bundle
        while bundle.object_view == bundle.parent_attr:
            bundle = bundle.parent

        if update_parent and changed_object.__class__ != bundle._meta.model:
            object_view, name = bundle.get_initialized_view_and_name(
                bundle.object_view, kwargs=self.kwargs
            )

            changed_object = object_view.get_object()
            changed_object.save()

        if not section:
            section = ""

        if url:
            url = urllib.parse.urlparse(url).path

        rep = str(instance)
        if rep:
            rep = rep[:255]

        log = CMSLog(
            action=action,
            url=url,
            section=section,
            model_repr=instance._meta.verbose_name,
            object_repr=rep,
            user_name=self.request.user.username,
            action_date=action_date,
        )
        log.save()

    def get_filter(self, **filter_kwargs):
        """
        Returns a list of Q objects that can be passed
        to an queryset for filtering.

        Default implementation returns a Q
        object for `base_filter_kwargs` and any
        passed in keyword arguments.
        """
        filter_kwargs.update(self.base_filter_kwargs)
        if filter_kwargs:
            return [models.Q(**filter_kwargs)]
        return []

    def get_queryset(self, **filter_kwargs):
        """
        Get the list of items for this view. This will
        call the `get_parent_object` method before doing
        anything else to ensure that a valid parent object
        is present. If a parent_object is returned it gets
        set to `self.parent_object`.

        If a queryset has been set then that queryset will be used.
        Otherwise the default manager for the provided
        model will be used.

        Once we have a queryset, the `get_filter` method
        is called and added to the queryset which is then
        returned.
        """
        self.parent_object = self.get_parent_object()

        if self.queryset is not None:
            queryset = self.queryset
            if hasattr(queryset, "_clone"):
                queryset = queryset._clone()
        elif self.model is not None:
            queryset = self.model._default_manager.filter()
        else:
            raise ImproperlyConfigured(
                f"'{self.__class__.__name__}' must define 'queryset' or 'model'"
            )

        q_objects = self.get_filter(**filter_kwargs)
        queryset = queryset.filter()
        for q in q_objects:
            queryset = queryset.filter(q)

        return queryset

    def get_parent_object(self):
        """
        Lookup a parent object. If parent_field is None
        this will return None. Otherwise this will try to
        return that object.

        The filter arguments are found by using the known url
        parameters of the bundle, finding the value in the url keyword
        arguments and matching them with the arguments in
        `self.parent_lookups`. The first argument in parent_lookups
        matched with the value of the last argument in the list of bundle
        url parameters, the second with the second last and so forth.

        For example let's say the parent_field attribute is 'gallery'
        and the current bundle knows about these url parameters:

        * adm_post
        * adm_post_gallery

        And the current value for 'self.kwargs' is:

        * adm_post = 2
        * adm_post_gallery = 3

        if parent_lookups isn't set the filter for the queryset
        on the gallery model will be:

        * pk = 3

        if parent_lookups is ('pk', 'post__pk') then the filter
        on the queryset will be:

        * pk = 3
        * post__pk = 2

        The model to filter on is found by finding the relationship
        in self.parent_field and filtering on that model.
        If a match is found, 'self.queryset` is changed to
        filter on the parent as described above and the parent
        object is returned. If no match is found, a Http404 error
        is raised.
        """

        if self.parent_field:
            # Get the model we are querying on
            if getattr(self.model._meta, "init_name_map", None):
                # pre-django-1.8
                cache = self.model._meta.init_name_map()
                field, mod, direct, m2m = cache[self.parent_field]
            else:
                field = self.model._meta.get_field(self.parent_field)
                m2m = field.is_relation and field.many_to_many
                direct = not field.auto_created or field.concrete

            to = None
            field_name = None
            if self.parent_lookups is None:
                self.parent_lookups = ("pk",)

            url_params = list(self.bundle.url_params)
            if url_params and getattr(self.bundle, "delegated", False):
                url_params = url_params[:-1]

            offset = len(url_params) - len(self.parent_lookups)
            kwargs = {}
            for i in range(len(self.parent_lookups) - 1):
                k = url_params[offset + i]
                value = self.kwargs[k]
                kwargs[self.parent_lookups[i + 1]] = value

            main_arg = self.kwargs[url_params[-1]]
            main_key = self.parent_lookups[0]

            if m2m:
                rel = getattr(self.model, self.parent_field)
                kwargs[main_key] = main_arg
                if direct:
                    to = rel.field.remote_field.to
                    field_name = self.parent_field
                else:
                    from django.db.models.fields.related import (
                        ForeignObjectRel
                    )

                    if isinstance(rel.rel, ForeignObjectRel):
                        to = rel.rel.related_model
                    else:
                        to = rel.rel.model
                    field_name = rel.rel.field.name
            else:
                to = field.remote_field.model
                if main_key == "pk":
                    try:
                        to_field = field.remote_field.field_name
                    except AttributeError:
                        to_field = field.remote_field.attname
                    if to_field == "vid":
                        to_field = "object_id"
                else:
                    to_field = main_key
                kwargs[to_field] = main_arg

            # Build the list of arguments
            try:
                obj = to.objects.get(**kwargs)
                if self.queryset is None:
                    if m2m:
                        self.queryset = getattr(obj, field_name)
                    else:
                        self.queryset = self.model.objects.filter(
                            **{self.parent_field: obj}
                        )
                return obj
            except to.DoesNotExist:
                raise http.Http404
        return None


class ModelCMSView(CMSView):
    """
    Base view for CMS views that interact with models.
    Inherits from CMSView.

    :param custom_model_name: A alternate verbose name for the \
    given model.
    """

    custom_model_name = None
    custom_model_name_plural = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pk_url_kwarg = None

    def write_message(self, status=messages.INFO, message=None):
        """
        Writes a message to django's messaging framework and
        returns the written message.

        :param status: The message status level. Defaults to \
        messages.INFO.
        :param message: The message to write. If not given, \
        defaults to appending 'saved' to the unicode representation \
        of `self.object`.
        """
        if not message:
            message = f"{self.object} saved"
        messages.add_message(self.request, status, message)
        return message

    def get_url_kwargs(self, request_kwargs=None, **kwargs):
        """
        If request_kwargs is not specified, self.kwargs is used instead.

        If 'object' is one of the kwargs passed. Replaces it with
        the value of 'self.slug_field' on the given object.
        """

        if not request_kwargs:
            request_kwargs = getattr(self, "kwargs", {})

        kwargs = super().get_url_kwargs(request_kwargs, **kwargs)
        obj = kwargs.pop("object", None)
        if obj:
            kwargs[self.slug_url_kwarg] = getattr(obj, self.slug_field, None)
        elif self.slug_url_kwarg in request_kwargs:
            kwargs[self.slug_url_kwarg] = request_kwargs[self.slug_url_kwarg]

        return kwargs

    def _model_name(self, plural=False):
        return helpers.model_name(
            self.model,
            self.custom_model_name,
            self.custom_model_name_plural,
            plural=plural,
        )

    @property
    def model_name_plural(self):
        """
        Property for getting the plural verbose name of the given model.
        Returns custom_model_name if present, otherwise returns
        the verbose_name of the model.
        """
        return self._model_name(plural=True)

    @property
    def model_name(self):
        """
        Property for getting the verbose name of the given model.
        Returns custom_model_name if present, otherwise returns
        the verbose_name of the model.
        """
        return self._model_name(plural=False)

    def get_render_data(self, **kwargs):
        """
        Adds the model_name to the context, then calls super.
        """
        kwargs["model_name"] = self.model_name
        kwargs["model_name_plural"] = self.model_name_plural
        return super().get_render_data(**kwargs)



from django.db import models

from scarlet.versioning import fields
from scarlet.versioning.models import (
    VersionView,
    Cloneable,
    BaseModel,
    BaseVersionedModel,
)


class Harmless:
    is_harmless = True


class ConcreteModel(models.Model):
    name = models.CharField(max_length=255)


class NameModel(models.Model):
    name = models.CharField(max_length=255)

    class Meta:
        abstract = True


class Abstract(BaseVersionedModel, NameModel):
    associates = fields.M2MFromVersion("self", blank=True)

    class Meta:
        abstract = True


class Author(VersionView, Abstract):
    def __str__(self):
        return self.name


class Cartoon(VersionView, NameModel):
    author = models.ForeignKey(
        Author, related_name="works", blank=True, null=True, on_delete=models.SET_NULL
    )

    def __str__(self):
        return self.name


class AbstractM2MBook(models.Model):
    books = fields.M2MFromVersion("Book", blank=True)
    cartoon = fields.FKToVersion(
        Cartoon, blank=True, null=True, on_delete=models.SET_NULL
    )

    class Meta:
        abstract = True


class Gallery(Cloneable):
    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


class Book(VersionView, NameModel, Harmless):
    _clone_related = ["review", "galleries"]

    author = models.ForeignKey(Author, on_delete=models.CASCADE)
    galleries = fields.M2MFromVersion(Gallery)

    def __str__(self):
        return self.name


class BookNoRelated(VersionView, NameModel, Harmless):
    author = models.ForeignKey(Author, on_delete=models.CASCADE)
    galleries = fields.M2MFromVersion(Gallery)

    def __str__(self):
        return self.name


class Review(Cloneable):
    book = fields.FKToVersion(Book, on_delete=models.CASCADE)
    text = models.CharField(max_length=255)

    def __str__(self):
        return self.text


class Store(VersionView, NameModel, AbstractM2MBook):
    pass


class NoReverse(VersionView, NameModel):
    _clone_related = ("rm2m",)


class RM2M(Cloneable):
    no = models.ManyToManyField(NoReverse)


class Image(Cloneable):
    name = models.CharField(max_length=255)
    cartoons = models.ManyToManyField(Cartoon)

    def __str__(self):
        return self.name


class CustomModel(BaseModel):
    reg_number = models.CharField(max_length=20)


class Gun(VersionView):
    name = models.CharField(max_length=20)

    class Meta:
        _base_model = CustomModel



import datetime

from django.test import TestCase
from django.utils import timezone, formats
from django.core.exceptions import ValidationError

from scarlet.versioning import manager

from . import models


class ModelTests(TestCase):
    fixtures = ("test_data.json",)

    def setUp(self):
        manager.activate("draft")

    def tearDown(self):
        manager.deactivate()

    def _check_different_book_versions(self, bd, bp):
        self.assertEqual(bd.object_id, bp.object_id)
        self.assertEqual(bd.id, bp.id)
        self.assertEqual(bd.pk, bp.pk)
        self.assertNotEqual(bd.vid, bp.vid)

    def _prepare_books(self):
        book = models.Book.objects.get(vid=1)
        book.publish()

        with manager.SwitchSchema("published"):
            bp = models.Book.normal.get(object_id=book.object_id)
        with manager.SwitchSchema("draft"):
            bd = models.Book.normal.get(object_id=book.object_id)

        return bp, bd

    def testNew(self):
        author = models.Author(name="MoYan")
        author.save()
        a = models.Author.objects.get(name="MoYan")
        self.assertTrue(a.pk)
        self.assertEqual(a.state, models.Author.DRAFT)

    def testCloneNoRelated(self):
        """
        class Cloneable _clone method.
        a clone has identical properties except vid
        """
        self.assertFalse(hasattr(models.BookNoRelated, "_clone_related"))

        book = models.BookNoRelated.objects.get(pk=1)
        last_save = book.last_save

        with manager.SwitchSchema("public"):
            n_books = models.BookNoRelated.objects.all().count()

        book._clone()
        book = models.BookNoRelated.objects.get(vid=1)

        with manager.SwitchSchema("public"):
            n_books_new = models.BookNoRelated.objects.all().count()

        self.assertEqual(n_books, n_books_new - 1)

        self.assertEqual(book.pk, 1)
        self.assertEqual(last_save, book.last_save)

        clone = models.BookNoRelated.objects.get(vid=2)
        self.assertEqual(clone.author, book.author)
        self.assertEqual(clone.object_id, book.object_id)

    def testCloneRelated(self):
        """
        attributes registered in _clone_related should be cloned as well
        """
        self.assertTrue(hasattr(models.Book, "_clone_related"))

        book = models.Book.objects.get(vid=1)
        n_review = models.Review.objects.all().count()
        self.assertEqual(n_review, 2)
        n_gallery = models.Gallery.objects.all().count()
        self.assertEqual(n_gallery, 2)

        book._clone()
        clone = models.Book.objects.get(vid__gt=1)
        book = models.Book.objects.get(vid=1)

        n_review = clone.review_set.all().count()
        self.assertEqual(n_review, 2)

        n_review = models.Review.objects.all().count()
        self.assertEqual(n_review, 4)

        n_gallery = models.Gallery.objects.all().count()
        self.assertEqual(n_gallery, 4)

    def testDeleteRelated(self):
        """
        related objects should be deleted as well
        """
        self.assertEqual(models.Book.objects.all().count(), 1)
        book = models.Book.objects.get(vid=1)

        self.assertEqual(models.Gallery.objects.all().count(), 2)
        self.assertEqual(models.Review.objects.all().count(), 2)

        book.delete()
        self.assertEqual(models.Book.objects.all().count(), 0)

        self.assertEqual(models.Gallery.objects.all().count(), 0)
        self.assertEqual(models.Review.objects.all().count(), 0)

    def testDeleteReverseM2M(self):
        no = models.NoReverse(name="test")
        no.save()
        with self.assertRaises(AssertionError):
            no.delete()

    def testDeleteVersion(self):
        """
        delete one version of an object should also delete all other versions
        """
        bp, bd = self._prepare_books()

        bp.delete()
        with manager.SwitchSchema("public"):
            self.assertFalse(models.Book.normal.filter(object_id=bd.object_id).exists())

        self.assertEqual(models.Gallery.objects.all().count(), 0)

    def testReverseM2MDeleteVersion(self):
        cartoon = models.Cartoon.objects.get(vid=1)
        cartoon.publish()

        with manager.SwitchSchema("public"):
            self.assertEqual(models.Cartoon.normal.all().count(), 2)

        klass = cartoon.get_version_class()
        klass.normal.filter(vid__gt=1).delete()

        self.assertEqual(models.Cartoon.normal.all().count(), 1)
        image = models.Image(pk=1)
        self.assertEqual(image.cartoons.all().count(), 1)

    def testDeleteWithM2M(self):
        cartoon = models.Cartoon.objects.get(vid=1)
        cartoon.publish()

        with manager.SwitchSchema("public"):
            self.assertEqual(models.Cartoon.normal.all().count(), 2)

            image = models.Image(pk=1)
            self.assertEqual(image.cartoons.all().count(), 1)

            cartoon.delete()
            self.assertEqual(image.cartoons.all().count(), 0)

    def testChangeRelated(self):
        """
        change related items of an object version won't affect other versions
        """
        bp, bd = self._prepare_books()

        with manager.SwitchSchema("draft"):
            bd_reviews = list(bd.review_set.all())
            bd_gallery = list(bd.galleries.all())

        # change related object
        bd_reviews[0].text = "changed"
        bd_reviews[0].save()
        bd_gallery[0].name = "changed"
        bd_gallery[0].save()

        self.assertTrue(bd.review_set.filter(text="changed").exists())
        self.assertTrue(bd.galleries.filter(name="changed").exists())

        with manager.SwitchSchema("published"):
            self.assertEqual(bp.review_set.all().count(), 2)
            self.assertEqual(bp.galleries.all().count(), 2)
            self.assertFalse(bp.review_set.filter(text="changed").exists())
            self.assertFalse(bp.galleries.filter(name="changed").exists())

    def testAddRelated(self):
        """
        add related item to an object version won't affect other versions
        """
        bp, bd = self._prepare_books()
        # add relationship
        g = models.Gallery(name="newly added")
        g.save()
        bp.galleries.add(g)

        with manager.SwitchSchema("published"):
            self.assertTrue(bp.galleries.filter(name="newly added").exists())
            self.assertEqual(bp.galleries.all().count(), 3)

        with manager.SwitchSchema("draft"):
            self.assertFalse(bd.galleries.filter(name="newly added").exists())
            self.assertEqual(bd.galleries.all().count(), 2)

        self.assertEqual(models.Gallery.objects.all().count(), 5)

    def testRelatedName(self):
        author = models.Author.objects.get(vid=1)
        self.assertTrue(hasattr(author, "works_version"))

    def testPublishAuthor(self):
        """
        publish is actually, make a clone of draft, and change the state
        of the clone to PUBLISHED
        """
        author = models.Author.objects.get(vid=1)
        self.assertEqual(author.state, models.Author.DRAFT)
        old_vid = author.vid

        author.publish()
        self.assertEqual(author.state, models.Author.PUBLISHED)
        self.assertNotEqual(author.vid, old_vid)

        author = models.Author.objects.get(vid=old_vid)
        self.assertEqual(author.state, models.Author.DRAFT)

    def testRegisterRelated(self):
        original = models.Book._clone_related
        setattr(models.Book, "_clone_related", [])
        models.Book.register_related("test")
        related = models.Book._clone_related
        self.assertEqual(related, ["test"])

        models.Book._clone_related = ("test",)
        models.Book.register_related("test2")
        related = models.Book._clone_related
        self.assertEqual(related, ["test", "test2"])

        setattr(models.Book, "_clone_related", original)

    def testGetVersion(self):
        author = models.Author.objects.get(pk=1)
        self.assertEqual(author.get_version(state=models.Author.PUBLISHED), None)

    def testPublishBook(self):
        from django.contrib.auth.models import User

        u = User.objects.create(username="admin")

        book = models.Book.objects.get(vid=1)
        book.publish(user=u)
        bd = models.Book.objects.get(object_id=book.object_id, state=models.Book.DRAFT)
        bp = book

        self.assertEqual(bp.user_published, "admin")

        self._check_different_book_versions(bd, bp)

    def testUnPublish(self):
        book = models.Book.objects.get(vid=1)
        klass = book.get_version_class()
        book.publish()
        self.assertEqual(klass.normal.filter().count(), 2)
        self.assertTrue(book.is_published)
        self.assertEqual(book.state, models.Book.PUBLISHED)

        book = models.Book.objects.get(vid=1)
        book.unpublish()
        self.assertEqual(klass.normal.filter().count(), 2)
        self.assertFalse(book.is_published)

    def testUnPublishWithScheduled(self):
        book = models.Book.objects.get(vid=1)
        klass = book.get_version_class()

        # schedule to publish
        t = timezone.now() + datetime.timedelta(days=7)
        book.publish(when=t)
        self.assertEqual(klass.normal.all().count(), 2)

        book = models.Book.objects.get(vid=1)
        book.unpublish()
        # make sure scheduled items get deleted
        self.assertFalse(
            klass.normal.filter(
                object_id=book.object_id, state=models.Book.SCHEDULED
            ).exists()
        )
        self.assertEqual(klass.normal.all().count(), 1)

    def testPublishWithPublishedAndScheduled(self):
        book = models.Book.objects.get(vid=1)
        klass = book.get_version_class()

        # publish the draft
        book.publish()
        self.assertEqual(klass.normal.all().count(), 2)
        self.assertTrue(klass.normal.filter(state=models.Book.PUBLISHED).exists())

        # schedule the draft to be published in future
        book = models.Book.objects.get(vid=1)
        t = timezone.now() + datetime.timedelta(days=20)
        book.publish(when=t)
        self.assertEqual(klass.normal.all().count(), 3)
        self.assertTrue(klass.normal.filter(state=models.Book.SCHEDULED).exists())

        # try publish the draft again; it will archive the item published above
        # and publish the current one
        book = models.Book.objects.get(vid=1)
        book.publish(when=timezone.now())
        self.assertEqual(klass.normal.all().count(), 4)
        self.assertTrue(klass.normal.filter(state=models.Book.ARCHIVED).exists())

    def testPublishWithScheduledOnly(self):
        book = models.Book.objects.get(vid=1)
        klass = book.get_version_class()

        # schedule the draft to be published in future
        book = models.Book.objects.get(vid=1)
        t = timezone.now() + datetime.timedelta(days=7)
        book.publish(when=t)
        self.assertEqual(klass.normal.all().count(), 2)
        self.assertTrue(klass.normal.filter(state=models.Book.SCHEDULED).exists())

        # the time stamp info in draft and scheduled items are identical
        draft = models.Book.objects.get(vid=1)
        self.assertEqual(draft.last_scheduled, book.last_scheduled)
        self.assertEqual(draft.date_published, t)
        self.assertEqual(book.date_published, t)

        # try publish the draft again; it should overwrite
        # scheduled item above and publish it now
        book = models.Book.objects.get(vid=1)
        book.publish()
        self.assertEqual(klass.normal.all().count(), 2)
        self.assertTrue(klass.normal.filter(state=models.Book.PUBLISHED).exists())

    def testMakeDraft(self):
        book = models.Book.objects.get(vid=1)
        self.assertEqual(book.name, "Book1")

        # draft cannot make draft -- return nothing inside make_draft function.
        with self.assertRaises(AssertionError):
            book.make_draft()

        klass = book.get_version_class()
        book = klass.normal.get(vid=1)
        book.make_draft()
        self.assertEqual(
            models.Book.objects.filter(object_id=book.object_id).count(), 1
        )

        book = klass.normal.get(vid=1)
        book.publish()

        # prepare a draft for deleting when calling make_draft
        draft = klass.normal.get(object_id=book.object_id, state=models.Book.DRAFT)
        self.assertEqual(draft.last_scheduled, draft.last_save)
        draft.name = "book2"
        draft.save()
        self.assertTrue(klass.normal.filter(name="book2").exists())

        # need to get the object again to get updated last_save value, coz the
        # one cached by django is not updated.
        draft = klass.normal.get(object_id=book.object_id, state=models.Book.DRAFT)
        self.assertEqual(draft.last_scheduled, book.last_scheduled)
        self.assertTrue(draft.last_scheduled < draft.last_save)

        book = klass.normal.get(object_id=book.object_id, state=models.Book.PUBLISHED)
        book.make_draft()
        # make sure the old draft get deleted
        self.assertFalse(klass.normal.filter(name="book2").exists())

        # make sure the underlying two versions of book are "same"
        manager.deactivate()
        bd = models.Book.objects.get(object_id=book.object_id, state=models.Book.DRAFT)
        bp = models.Book.objects.get(
            object_id=book.object_id, state=models.Book.PUBLISHED
        )
        manager.activate("draft")

        self.assertEqual(bd.name, "Book1")
        self._check_different_book_versions(bd, bp)

    def testRevert(self):
        """
        update the draft; revert to the archived;
        the update should be abolished.
        """
        obp, obd = self._prepare_books()
        obd.publish()
        klass = obp.get_version_class()

        with manager.SwitchSchema("public"):
            bp = klass.normal.get(object_id=obp.object_id, state=models.Book.PUBLISHED)
            bd = klass.normal.get(object_id=obp.object_id, state=models.Book.DRAFT)
            ba = klass.normal.get(object_id=obp.object_id, state=models.Book.ARCHIVED)

        bd.name = "the sea"
        bd.save()
        self.assertEqual(bd.name, "the sea")
        self.assertEqual(bp.name, "Book1")
        self.assertEqual(ba.name, "Book1")

        ba.make_draft()
        with manager.SwitchSchema("public"):
            self.assertEqual(models.Book.normal.all().count(), 3)
            bp2 = models.Book.normal.get(
                object_id=obp.object_id, state=models.Book.PUBLISHED
            )
            bd2 = models.Book.normal.get(
                object_id=obp.object_id, state=models.Book.DRAFT
            )
            ba2 = models.Book.normal.get(
                object_id=obp.object_id, state=models.Book.ARCHIVED
            )

        # old draft gets deleted,
        self.assertFalse(models.Book.objects.filter(vid=1).exists())
        # new draft generated
        self.assertEqual(bd2.name, "Book1")
        # other versions stay same
        self.assertEqual(bp2.name, "Book1")
        self.assertEqual(ba2.name, "Book1")

    def testDraftRelated(self):
        """
        get draft, and published. add new gallery item to draft, and publish
        it then revert back.

        remove gallery item from draft; publish it; revert back;
        """
        obp, obd = self._prepare_books()
        klass = obp.get_version_class()

        bp = klass.normal.get(object_id=obp.object_id, state=models.Book.PUBLISHED)
        bd = klass.normal.get(object_id=obp.object_id, state=models.Book.DRAFT)

        self.assertEqual(models.Gallery.objects.all().count(), 4)
        self.assertEqual(bd.vid, 1)

        # change draft
        g = models.Gallery(name="old man")
        g.save()
        bd.galleries.add(g)
        self.assertEqual(models.Gallery.objects.all().count(), 5)

        # change should occur only in draft
        self.assertEqual(bp.galleries.all().count(), 2)
        self.assertFalse(bp.galleries.filter(name="old man").exists())
        self.assertEqual(bd.galleries.all().count(), 3)
        self.assertTrue(bd.galleries.filter(name="old man").exists())

        # publish the changed draft; change should now in published version
        bd.publish()
        bp2 = klass.normal.get(object_id=obp.object_id, state=models.Book.PUBLISHED)
        klass.normal.filter(object_id=obp.object_id, state=models.Book.DRAFT).exists()
        ba2 = klass.normal.get(object_id=obp.object_id, state=models.Book.ARCHIVED)
        self.assertEqual(models.Gallery.objects.all().count(), 8)
        self.assertEqual(bp2.galleries.all().count(), 3)
        self.assertTrue(bp2.galleries.filter(name="old man").exists())
        # meanwhile, published get archived
        self.assertEqual(ba2.vid, bp.vid)

        # revert archived back to draft
        ba2.make_draft()
        self.assertEqual(klass.normal.all().count(), 3)
        self.assertEqual(models.Gallery.objects.all().count(), 7)
        bp3 = klass.normal.get(object_id=obp.object_id, state=models.Book.PUBLISHED)
        bd3 = klass.normal.get(object_id=obp.object_id, state=models.Book.DRAFT)
        ba3 = klass.normal.get(object_id=obp.object_id, state=models.Book.ARCHIVED)
        self.assertEqual(bd3.galleries.all().count(), 2)
        self.assertFalse(bd3.galleries.filter(name="old man").exists())

        # remove one gallery item from draft
        g = bd3.galleries.filter(name="Gallery1")
        self.assertEqual(len(g), 1)
        bd3.galleries.remove(g[0])
        self.assertEqual(models.Gallery.objects.all().count(), 7)
        self.assertEqual(bd3.galleries.all().count(), 1)
        self.assertEqual(bp3.galleries.all().count(), 3)
        self.assertEqual(ba3.galleries.all().count(), 2)

        # publish the draft, we now have 1 published, 1 draft, and 2 archived
        bd3.publish()
        self.assertEqual(klass.normal.all().count(), 4)
        self.assertEqual(models.Gallery.objects.all().count(), 8)
        bp4 = klass.normal.get(object_id=obp.object_id, state=models.Book.PUBLISHED)
        bd4 = klass.normal.get(object_id=obp.object_id, state=models.Book.DRAFT)
        ba4 = klass.normal.get(object_id=obp.object_id, vid=bp3.vid)
        # now both the draft and published should have only 1 gallery item
        self.assertEqual(bp4.galleries.all().count(), 1)
        self.assertFalse(bp4.galleries.filter(name="Gallery1").exists())
        self.assertEqual(bd4.galleries.all().count(), 1)
        self.assertFalse(bd4.galleries.filter(name="Gallery1").exists())

        # revert back to the new archived status (ba4 is actually bp3)
        ba4.make_draft()
        self.assertEqual(klass.normal.all().count(), 4)
        self.assertEqual(models.Gallery.objects.all().count(), 10)

        bp5 = klass.normal.get(object_id=obp.object_id, state=models.Book.PUBLISHED)
        bd5 = klass.normal.get(object_id=obp.object_id, state=models.Book.DRAFT)
        # now the draft should havd 3 gallery items
        self.assertEqual(bd5.galleries.all().count(), 3)
        self.assertTrue(bd5.galleries.filter(name="Gallery1").exists())
        self.assertTrue(bd5.galleries.filter(name="old man").exists())
        # published is not affected
        self.assertEqual(bp5.galleries.all().count(), 1)
        self.assertTrue(bp5.galleries.filter(name="Gallery2").exists())

    def testMakeDraftWithoutDraft(self):
        # it is ok for published book to make draft, when there is no draft.
        book = models.Book.objects.get(vid=1)
        klass = book.get_version_class()
        book = klass.normal.get(vid=1)

        book.publish()
        draft = klass.normal.get(object_id=book.object_id, state=models.Book.DRAFT)
        draft.delete()

        self.assertTrue(
            klass.normal.filter(
                object_id=book.object_id, state=models.Book.PUBLISHED
            ).exists()
        )
        book.make_draft()
        self.assertEqual(klass.normal.all().count(), 2)

    def testPurgeArchive(self):
        # create a few ARCHIVED versions
        for i in range(0, 10):
            book = models.Book.objects.get(vid=1)
            book.publish()

        # the underlying versions
        klass = book.get_version_class()
        self.assertEqual(klass.normal.filter(state=models.Book.ARCHIVED).count(), 9)
        self.assertEqual(klass.normal.filter(state=models.Book.DRAFT).count(), 1)
        self.assertEqual(klass.normal.filter(state=models.Book.PUBLISHED).count(), 1)

        # make sure only NUM_KEEP_ARCHIVED (5) number of ARCHIVED are kept
        book.purge_archives()
        self.assertEqual(klass.normal.filter(state=models.Book.ARCHIVED).count(), 5)
        self.assertEqual(models.Gallery.objects.all().count(), 14)

    def testStatus1(self):
        book = models.Book.objects.get(vid=1)
        self.assertEqual(book.state, models.Book.DRAFT)
        self.assertEqual(book.status_line(), "Draft saved: 08/30/2012")

        # schedule to publish in future. should get status as Scheduled.
        book.publish(when=timezone.now() + datetime.timedelta(days=7))
        draft = models.Book.objects.get(vid=1)

        self.assertEqual(draft.last_save, draft.last_scheduled)
        self.assertFalse(draft.v_last_save)  # v_last_save empty
        status = (
            "Publish Scheduled: "
            f"{formats.date_format(draft.date_published, 'SHORT_DATE_FORMAT')}"
        )
        self.assertEqual(status, draft.status_line())

        # publish it now. should get status as Published
        when = timezone.now()
        book = models.Book.objects.get(vid=1)
        book.publish(when=when)
        draft2 = models.Book.objects.get(vid=1)

        self.assertEqual(draft2.last_save, draft2.last_scheduled)
        self.assertEqual(draft2.last_save, draft2.v_last_save)
        status = (
            "Published: "
            f"{formats.date_format(draft2.date_published, 'SHORT_DATE_FORMAT')}"
        )
        self.assertEqual(status, draft2.status_line())
        self.assertEqual(draft2.date_published, when)

        # schedule to publish in future again. should get status as Scheduled
        book = models.Book.objects.get(vid=1)
        book.publish(when=timezone.now() + datetime.timedelta(days=7))
        draft = models.Book.objects.get(vid=1)
        self.assertEqual(draft.last_save, draft.last_scheduled)
        self.assertTrue(draft.last_scheduled > draft.v_last_save)
        status = (
            "Publish Scheduled: "
            f"{formats.date_format(draft.date_published, 'SHORT_DATE_FORMAT')}"
        )
        self.assertEqual(status, draft.status_line())

    def testStatus2(self):
        book = models.Book.objects.get(vid=1)
        self.assertEqual(book.state, models.Book.DRAFT)
        self.assertEqual(book.status_line(), "Draft saved: 08/30/2012")

        book.publish()
        draft = models.Book.objects.get(vid=1)
        self.assertEqual(book.v_last_save, draft.v_last_save)
        self.assertEqual(draft.last_save, draft.last_scheduled)
        self.assertEqual(draft.last_save, draft.v_last_save)
        status = (
            "Published: "
            f"{formats.date_format(draft.last_save, 'SHORT_DATE_FORMAT')}"
        )
        self.assertEqual(status, book.status_line())

    def testSwitchSchema(self):
        book = models.Book.objects.get(vid=1)
        book.publish()
        self.assertEqual(models.Book.normal.all().count(), 1)
        with manager.SwitchSchema("public"):
            self.assertEqual(models.Book.normal.all().count(), 2)

        self.assertEqual(models.Book.normal.all().count(), 1)
        manager.deactivate()
        self.assertEqual(models.Book.normal.all().count(), 2)

    def testVersionedUnique(self):
        """
        use versioned_unique to force unique on field value
        """
        self.assertFalse(hasattr(models.Author, "versioned_unique"))
        setattr(models.Author, "versioned_unique", ["name"])
        self.assertEqual(models.Author.versioned_unique, ["name"])
        a1 = models.Author.objects.get(vid=1)

        # should fail on second(duplicate) instance
        a2 = models.Author(name=a1.name)
        with self.assertRaises(ValidationError):
            a2.validate_unique()

        # multiple versions are not affected (can co-exists)
        klass = a1.get_version_class()
        self.assertEqual(klass.normal.filter(object_id=a1.object_id).count(), 1)
        a1.publish()
        self.assertEqual(klass.normal.filter(object_id=a1.object_id).count(), 2)
        self.assertTrue(
            klass.normal.filter(object_id=a1.object_id, state=a1.DRAFT).exists()
        )
        self.assertTrue(
            klass.normal.filter(object_id=a1.object_id, state=a1.PUBLISHED).exists()
        )
        a1.validate_unique()

    def testCustomBaseModel(self):
        """
        custom base_model; defined by _base_model in Meta;
        should turn on 'should_save_base' to save fields in _base_model.
        """
        gun = models.Gun()
        self.assertEqual(gun._meta._base_model, models.CustomModel)

        # should have all the fields, including the one from _base_model
        self.assertTrue(hasattr(gun, "created_date"))
        self.assertTrue(hasattr(gun, "v_last_save"))
        self.assertTrue(hasattr(gun, "reg_number"))
        self.assertTrue(hasattr(gun, "is_published"))
        self.assertFalse(hasattr(gun, "none_exist_field"))

        self.assertTrue(hasattr(gun, "should_save_base"))

        self.assertEqual(gun.should_save_base, False)

        # should be created normally
        gun.name = "AK47"
        gun.save()

        gun = models.Gun.objects.get(name="AK47")
        self.assertEqual(gun.object_id, 1)
        self.assertFalse(gun.reg_number)

        self.assertEqual(gun.should_save_base, False)
        gun.reg_number = "A1234"
        gun.save()
        gun = models.Gun.objects.get(vid=1)
        # not get saved
        self.assertFalse(gun.reg_number)

        gun.should_save_base = True
        gun.reg_number = "B3210"
        gun.save()
        gun = models.Gun.objects.get(vid=1)
        # should get saved now, since should_save_base is on
        self.assertEqual(gun.reg_number, "B3210")


class TransactionSwitchTests(TestCase):
    fixtures = ("test_data.json",)

    def setUp(self):
        manager.activate("draft")

    def tearDown(self):
        manager.deactivate()

    def testDeleteWithNormal(self):
        author = models.Author.objects.get(vid=1)
        author_id = author.pk
        author.publish()

        cartoon = models.Cartoon.objects.get(vid=1)
        self.assertEqual(cartoon.author, author)
        cartoon.publish()

        book = models.Book.objects.get(vid=1)
        book.publish()

        book2 = models.Book(name="test", author_id=2)
        book2.save()

        with manager.SwitchSchema("public"):
            self.assertEqual(models.Author.normal.filter(associates__pk=2).count(), 2)
            self.assertEqual(models.Cartoon.normal.filter(author=author).count(), 2)
            self.assertEqual(
                models.Cartoon.normal.filter(author__isnull=True).count(), 0
            )

        # Switch to draft mode to restrict what the ORM
        # will normally see
        with manager.SwitchSchema("draft"):
            self.assertEqual(models.Book.objects.filter(author=author).count(), 1)
            self.assertEqual(
                models.Gallery.objects.filter(book__author=author).count(), 2
            )
            self.assertEqual(
                models.Review.objects.filter(book__author=author).count(), 2
            )

            book = models.Book.objects.get(vid=1)
            self.assertEqual(book.author, author)
            author.delete()

        with manager.SwitchSchema("public"):
            self.assertEqual(models.Author.normal.filter(associates__pk=2).count(), 0)
            self.assertFalse(models.Author.normal.filter(pk=author_id).exists())
            self.assertEqual(
                models.Cartoon.normal.filter(author_id=author_id).count(), 0
            )
            self.assertEqual(
                models.Cartoon.normal.filter(author__isnull=True).count(), 2
            )
            self.assertEqual(models.Book.normal.all().count(), 1)
            self.assertEqual(models.Review.objects.all().count(), 0)
            self.assertEqual(models.Gallery.objects.all().count(), 0)




from django.conf import settings

from . import groups

BASE_GROUPS = getattr(settings, "ACCOUNTS_BASE_GROUPS", (groups.ADMIN,))

ACCOUNTS_FORBIDDEN_USERNAMES = getattr(
    settings,
    "ACCOUNTS_FORBIDDEN_USERNAMES",
    ("signup", "signout", "signin", "activate", "me", "password"),
)



from django.db import models
from django.utils import timezone


class CMSLog(models.Model):
    """
    A log of changes that have happened in the cms.
    Sets constants for the following actions:

    * SAVE
    * DELETE
    * PUBLISH
    * UNPUBLISH
    * SCHEDULE
    """

    SAVE = 0
    DELETE = 2
    PUBLISH = 3
    UNPUBLISH = 4
    SCHEDULE = 5

    ACTIONS = (
        (SAVE, "Save"),
        (DELETE, "Delete"),
        (PUBLISH, "Published"),
        (UNPUBLISH, "Unpublished"),
        (SCHEDULE, "Scheduled"),
    )

    action = models.PositiveIntegerField(choices=ACTIONS)
    action_date = models.DateTimeField(blank=True, null=True)

    model_repr = models.CharField(max_length=255)
    object_repr = models.CharField(max_length=255)
    url = models.CharField(max_length=255, blank=True)
    section = models.CharField(max_length=255, blank=True)

    user_name = models.CharField(max_length=255)
    when = models.DateTimeField(default=timezone.now)

    class Meta:
        app_label = "cms"



""" This code provides a decorator / context manager for transaction management
    in Django on PostgreSQL.  It is intended as a replacement for the existing
    Django commit_on_success() function, and provides some nice features:

    * Nested transactions: The top-level transaction will be a
      BEGIN/COMMIT/ROLLBACK block; inner "transactions" are implemented as
      savepoints.
    * Commits even if is_dirty is False, eliminating the mistake of forgetting
      to set the dirty flag when doing database-modifying raw SQL.
    * Better interaction with pgPool II, if you're using it.
    * A workaround for a subtle but nasty bug in Django's transaction
      management.

    For full details, check the README.md file.
"""

from functools import wraps

import psycopg2.extensions
from django.db import connections, DEFAULT_DB_ALIAS, transaction


class _Transaction:

    """
    This class manages a particular transaction or savepoint block, using
    context manager-style __enter__ and __exit__ statements.  We don't use it
    directly (for reasons noted below), but as a delegate for the
    _TransactionWrapper class.
    """

    def __init__(self, using):
        self.using = using
        self.sid = None

    def __enter__(self):
        if transaction.is_managed(self.using):
            # We're already in a transaction; create a savepoint.
            self.sid = transaction.savepoint(self.using)
        else:
            transaction.enter_transaction_management(using=self.using)
            transaction.managed(True, using=self.using)

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value is None:
            # commit operation
            if self.sid is None:
                # Outer transaction
                try:
                    transaction.commit(self.using)
                except Exception:
                    transaction.rollback(self.using)
                    raise
                finally:
                    self._leave_transaction_management()
            else:
                # Inner savepoint
                try:
                    transaction.savepoint_commit(self.sid, self.using)
                except Exception:
                    transaction.savepoint_rollback(self.sid, self.using)
                    raise
        else:
            # rollback operation
            if self.sid is None:
                # Outer transaction
                transaction.rollback(self.using)
                self._leave_transaction_management()
            else:
                # Inner savepoint
                transaction.savepoint_rollback(self.sid, self.using)

        # Returning False here means we did not gobble up the exception, so the
        # exception process should continue.
        return False

    def _leave_transaction_management(self):
        transaction.leave_transaction_management(using=self.using)
        if (
            not connections[self.using].is_managed()
            and connections[self.using].features.uses_autocommit
        ):
            connections[self.using]._set_isolation_level(
                psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT
            )
            # Patch for bug in Django's psycopg2 backend; see:
            # https://code.djangoproject.com/ticket/16047


class _TransactionWrapper:

    """ This class wraps the above _Transaction class.  We do this to allow
        reentrancy and thread-safety.  When being used as a decorator, only one
        _TransactionWrapper object is created per function being wrapped, and
        thus we can't store the state of the tranasction here (because multiple
        concurrent calls in the same address space to the same function would
        cause the state to be crunched), so delegate that to a _Transaction
        object that is created at the appropriate time.

        The "appropriate time" is two places: If the _TransactionWrapper is
        being used as a context manager, it's when the __enter__ function is
        called; if it is being used as a decorator, it's when the decorated
        function is about to be called (see the `inner` function below in
        __call__).

        The __enter__ and __exit__ functions on _TransactionWrapper are only
        called if we're using xact() as a context manager; if we're using it as
        a decorator, they're skipped and self.transaction is always None.
        Similarly, __call__ is not used if this is a context manager usage.
        This is not super-elegant, but it's the only way I've found to allow
        xact() to be used as both a context manager and a decorator using the
        same syntax.
    """

    def __init__(self, using):
        self.using = using
        self.transaction = None

    def __enter__(self):
        if self.transaction is None:
            self.transaction = _Transaction(self.using)
        return self.transaction.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        return self.transaction.__exit__(exc_type, exc_value, traceback)

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            with _Transaction(self.using):
                return func(*args, **kwargs)

        return inner


def xact(using=None):
    if using is None:
        using = DEFAULT_DB_ALIAS
    if callable(using):
        # We end up here if xact is being used as a completely bare decorator:
        #   @xact
        # (not even an empty parameter list)
        return _TransactionWrapper(DEFAULT_DB_ALIAS)(using)
        # Note that `using` here is *not* the database alias;
        # it's the actual function
        # being decorated.
    else:
        # We end up here if xact is being used as a parameterized decorator
        # (including
        # default parameter):
        #    @xact(db)
        # or @xact()
        # ... or as a context manager:
        #    with xact():
        #       ...
        return _TransactionWrapper(using)


# -----------------------------------------------------------------------------
# This software is licensed under the PostgreSQL License:
#
#   http://www.postgresql.org/about/licence/
#
# Copyright (c) 2012 Christophe Pettus
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose, without fee, and without a written agreement is
# hereby granted, provided that the above copyright notice and this paragraph
# and the following two paragraphs appear in all copies.
#
# IN NO EVENT SHALL CHRISTOPHE PETTUS BE LIABLE TO ANY PARTY FOR DIRECT,
# INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST
# PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
# IF CHRISTOPHE PETTUS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#
# CHRISTOPHE PETTUS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS,
# AND CHRISTOPHE PETTUS HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
# SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.



from django.db import models

from . import handler


class AutoTagModel(models.Model):
    tags = handler.get_tag_manager()
    tags.verbose_name = "Search Tags"
    tags.help_text = "Search tags are used to prefilter assets when including an asset on your object"

    def _set_auto_tags(self, value):
        if value:
            self._pending_tags = set(value)

    def _get_auto_tags(self):
        if not getattr(self, "_pending_tags", None):
            self._pending_tags = set()
        return self._pending_tags

    auto_tags = property(_get_auto_tags, _set_auto_tags)

    def add_pending_tags(self, field_tags):
        if self.auto_tags:
            needed_tags = set(field_tags).union(self.auto_tags)
            existing_tags = {}
            for t in handler.get_model().objects.filter(name__in=needed_tags):
                existing_tags[t.name] = t

            current_tags = set([t.pk for t in self.tags.all()])
            for tag in needed_tags:
                t = existing_tags.get(tag)
                if not t:
                    t, created = handler.get_model().objects.get_or_create(name=tag)

                if t.pk not in current_tags:
                    self.tags.add(t)

    class Meta:
        abstract = True



from django.http import HttpResponse

from scarlet.cache.views import CacheView, CacheListView
from .cache_groups import cache_group, CACHE_KEY


class DummyCacheView(CacheView):
    def get_cache_version(self):
        return cache_group.get_version(CACHE_KEY)

    def get(self, request, *args, **kwargs):
        return HttpResponse("response")


class GetCacheVersionNotImplementedView(CacheView):
    def get(self, request, *args, **kwargs):
        return HttpResponse("get_cache_version not implemented")


class DummyCacheListView(CacheListView):
    def get_cache_version(self):
        return cache_group.get_version(CACHE_KEY)

    def get(self, request, *args, **kwargs):
        return HttpResponse("response")


class GetCacheVersionNotImplementedListView(CacheListView):
    def get(self, request, *args, **kwargs):
        return HttpResponse("get_cache_version not implemented")
